import collections
try:
    import cStringIO as StringIO
except ImportError:
    from six import StringIO
import logging
import math
import numpy as np
import os
from universe import pyprofile
import re
import struct
import zlib

from twisted.internet import defer, protocol, threads

from universe import utils
from universe.twisty import reactor
from universe.vncdriver import auth, constants, error, screen, server_messages

logger = logging.getLogger(__name__)

# pyprofile.profile.print_frequency = 1
# pyprofile.profile.print_filter = lambda event: event.startswith('vncdriver.recv_rectangle')

def peer_address(peer):
    return '{}:{}'.format(peer.host, peer.port)

class Framebuffer(object):
    def __init__(self, width, height, server_pixel_format, name):
        # self.observer = observer

        self.width = width
        self.height = height
        self.name = name

        self.numpy_screen = screen.NumpyScreen(width, height)
        self.apply_format(server_pixel_format)

    def apply_format(self, server_pixel_format):
        self.server_pixel_format = server_pixel_format
        (self.bpp, self.depth, self.bigendian, self.truecolor,
         self.redmax, self.greenmax, self.bluemax,
         self.redshift, self.greenshift, self.blueshift) = \
           struct.unpack('!BBBBHHHBBBxxx', server_pixel_format)

        # x11vnc will set truecolor == 0 by default. We just have to
        # hope the client will override it.
        #
        # assert self.bigendian == 0
        # Don't want to deal with colormaps...
        # assert self.truecolor == 1

        # Derived values
        self.bypp = self.bpp // 8 # bytes per pixel

        shifts = [self.redshift, self.greenshift, self.blueshift]
        assert set(shifts) == set([0, 8, 16]), 'Surprising pixelformat: {}'.format(self.__dict__)
        # How to cycle pixels from images to get RGB
        self.color_cycle = np.argsort(shifts)
        self.server_init = struct.pack('!HH16sI', self.width, self.height, self.server_pixel_format, len(self.name)) + self.name

        self.numpy_screen.color_cycle = self.color_cycle

class VNCClient(protocol.Protocol, object):
    def __init__(self):
        self.numpy_screen = None

        self.buf = []
        self.buf_len = 0

        self.initialized = False

        # These could be passed around in
        # expected_args/expected_kwargs, but they are needed in so
        # many places it'd become confusing.
        self._remaining_rectangles = None
        self._rectangles = None

        self._pause = False

        self.expect(self.recv_ProtocolVersion_Handshake, 12)

        self._close = False
        self.zlib_decompressor = zlib.decompressobj()

        self._pointer_x = None
        self._pointer_y = None

    def connectionLost(self, reason):
        if not self._close:
            logger.info('Server %s hung up: %s', peer_address(self.transport.getPeer()), reason)
            self._error(error.Error('[{}] Lost connection: {}'.format(self.factory.label, reason)))

    def sendMessage(self, data):
        pyprofile.incr('vnc_client.data.sent.messages')
        pyprofile.incr('vnc_client.data.sent.bytes', len(data), unit=pyprofile.BYTES)
        if self.transport:
            self.transport.write(data)

    def dataReceived(self, data):
        pyprofile.incr('vnc_client.data.received.messages')
        pyprofile.incr('vnc_client.data.received.bytes', len(data), unit=pyprofile.BYTES)

        self.buf.append(data)
        self.buf_len += len(data)
        logger.debug('Received data: %s bytes (brings us to %s total)', len(data), self.buf_len)

        self.flush()

    def flush(self):
        if self.buf_len < self.expected_len:
            return
        elif self._pause:
            # Not strictly needed, but short circuits in case we're
            # paused for a while.
            return
        elif self._close:
            return

        buffer = b''.join(self.buf)
        while len(buffer) >= self.expected_len:
            logger.debug('Remaining in buffer: %s bytes', len(buffer))
            if self._pause:
                logger.debug('Pausing with %s bytes left in the buffer', len(buffer))
                break
            block, buffer = buffer[:self.expected_len], buffer[self.expected_len:]
            if not self.handle(self.expected, block):
                logger.debug('Stopping due to error in handle()')
                buffer = block + buffer
                break

        self.buf[:] = [buffer]
        self.buf_len = len(buffer)

    def handle(self, type, block):
        logger.debug('Handling server event: type=%s', type.__name__)
        try:
            self.expected(block, *self.expected_args, **self.expected_kwargs)
            return True
        except Exception as e:
            self._error(e)
            return False

    def recv_ProtocolVersion_Handshake(self, block):
        match = re.search(b'^RFB (\d{3}).(\d{3})\n$', block)
        assert match, 'Expected RFB line, but got: {!r}'.format(block)
        major = int(match.group(1))
        minor = int(match.group(2))
        self.sendMessage(b'RFB 003.003\n')

        self.expect(self.recv_Security_Handshake, 4)

    def recv_Security_Handshake(self, block):
        (auth,) = struct.unpack('!I', block)
        if auth == 0:
            self.expect(self.recv_SecurityResult_Handshake_failed_length, 4)
        elif auth == 1:
            self.send_ClientInit()
        elif auth == 2:
            self.expect(self.recv_VNC_Authentication, 16)
        else:
            assert False, 'Bad auth: {}'.format(auth)

    def recv_SecurityResult_Handshake(self, block):
        (result,) = struct.unpack('!xxxB', block)
        if result == 0:
            logger.debug('VNC Auth succeeded')
            self.send_ClientInit()
        elif result == 1:
            logger.debug('VNC Auth failed.')
            # Server optionally can say why
            self.expect(self.recv_SecurityResult_Handshake_failed_length, 4)
        else:
            assert False, 'Bad security result: {}'.format(result)

    def recv_SecurityResult_Handshake_failed_length(self, block):
        (length,) = struct.unpack('!I', block)
        self.expect(self.recv_SecurityResult_Handshake_failed_reason, length)

    def recv_SecurityResult_Handshake_failed_reason(self, block):
        logger.info('Connection to server failed: %s', block)

    def recv_VNC_Authentication(self, block):
        response = auth.challenge_response(block)
        self.sendMessage(response)
        self.expect(self.recv_SecurityResult_Handshake, 4)

    def send_ClientInit(self):
        shared = True
        self.sendMessage(struct.pack('!B', shared))
        self.expect(self.recv_ServerInit, 24)

    def recv_ServerInit(self, block):
        # This can be rewritten, so don't proxy immediately
        (width, height, server_pixel_format, namelen) = struct.unpack('!HH16sI', block)
        self.expect(self.recv_ServerInit_name, namelen, width, height, server_pixel_format)

    def recv_ServerInit_name(self, block, width, height, server_pixel_format):
        self.framebuffer = Framebuffer(width, height, server_pixel_format, block)
        self.numpy_screen = self.framebuffer.numpy_screen

        self.initialized = True
        self.send_PixelFormat()
        self.send_SetEncodings([
            constants.ZRLE_ENCODING,
            constants.ZLIB_ENCODING,
            constants.RAW_ENCODING,
        ])
        self.send_FramebufferUpdateRequest(incremental=1)

        if self.factory.deferred:
            self.factory.deferred.callback(self)

        # All connected!
        self.expect(self.recv_ServerToClient, 1)

    def recv_ServerToClient(self, block):
        (message_type,) = struct.unpack('!B', block)
        if message_type == 0:
            # self.observer.server_data_block(block)
            self.expect(self.recv_FramebufferUpdate, 3)
        elif message_type == 1:
            # self.observer.server_data_block(block)
            self.expect(self.recv_SetColorMapEntries, 5)
        elif message_type == 2:
            # self.observer.server_data_block(block)
            # self.observer.server_bell()
            self.expect(self.recv_ServerToClient, 1)
        elif message_type == 3:
            # Just drop ServerCutText messages
            self.expect(self.recv_ServerCutText, 7)
        else:
            assert False, 'Unknown server to client message type received: {}'.format(message_type)

    def recv_FramebufferUpdate(self, block):
        (number_of_rectangles, ) = struct.unpack('!xH', block)
        logger.debug('Receiving %d rectangles', number_of_rectangles)
        pyprofile.incr('vncdriver.framebuffer_update')
        pyprofile.incr('vncdriver.framebuffer_update.number_of_rectangles', number_of_rectangles)
        self._remaining_rectangles = number_of_rectangles
        self._rectangles = []

        self._process_rectangles()

    def _process_rectangles(self):
        if self._remaining_rectangles > 0:
            self.expect(self.recv_Rectangle, 12)
        else:
            framebuffer_update = server_messages.FramebufferUpdate(self._rectangles)
            self.numpy_screen.apply(framebuffer_update)
            self._remaining_rectangles = None
            self._rectangles = None

            # Once you're done, send another framebufferUpdateRequest.
            #
            # Empirically, after a few framebufferUpdateRequest's without
            # any changes, Xvnc will hold off on sending another
            # framebuffer update until the display data changes.
            self.send_FramebufferUpdateRequest(incremental=1)
            self.expect(self.recv_ServerToClient, 1)

    def recv_Rectangle(self, block):
        self._remaining_rectangles -= 1

        (x, y, width, height, encoding) = struct.unpack('!HHHHi', block)

        if encoding == constants.RAW_ENCODING:
            self.expect(self.recv_DecodeRAW, width*height*self.framebuffer.bypp, x, y, width, height)
        elif encoding == constants.ZRLE_ENCODING:
            self.expect(self.recv_DecodeZRLE, 4, x, y, width, height)
        elif encoding == constants.ZLIB_ENCODING:
            self.expect(self.recv_DecodeZlib, 4, x, y, width, height)
        elif encoding == constants.PSEUDO_CURSOR_ENCODING:
            length = width * height * self.framebuffer.bypp
            length += int(math.floor((width + 7.0) / 8)) * height
            self.expect(self.recv_DecodePseudoCursor, length, x, y, width, height)
        else:
            assert False, 'Unknown pixel encoding received: {}'.format(encoding)

    # Encodings

    def recv_DecodePseudoCursor(self, block, x, y, width, height):
        # cf https://github.com/sibson/vncdotool/blob/master/vncdotool/rfb.py
        rectangle = server_messages.PseudoCursorEncoding.parse_rectangle(self, x, y, width, height, block)
        self._rectangles.append(rectangle)
        self._process_rectangles()

    def recv_DecodeRAW(self, block, x, y, width, height):
        pyprofile.incr('vncdriver.recv_rectangle.raw_encoding')
        pyprofile.incr('vncdriver.recv_rectangle.raw_encoding.bytes', len(block), unit=pyprofile.BYTES)
        rectangle = server_messages.RAWEncoding.parse_rectangle(self, x, y, width, height, block)
        self._rectangles.append(rectangle)
        self._process_rectangles()

    def recv_DecodeZRLE(self, block, x, y, width, height):
        pyprofile.incr('vncdriver.recv_rectangle.zrle_encoding')
        pyprofile.incr('vncdriver.recv_rectangle.zrle_encoding.bytes', len(block), unit=pyprofile.BYTES)

        (length,) = struct.unpack('!I', block)
        self.expect(self.recv_DecodeZRLE_value, length, x, y, width, height)

    def recv_DecodeZRLE_value(self, block, x, y, width, height):
        pyprofile.incr('vncdriver.recv_rectangle.zrle_encoding.bytes', len(block), unit=pyprofile.BYTES)
        rectangle = server_messages.ZRLEEncoding.parse_rectangle(self, x, y, width, height, block)
        self._rectangles.append(rectangle)
        self._process_rectangles()

    def recv_DecodeZlib(self, block, x, y, width, height):
        pyprofile.incr('vncdriver.recv_rectangle.zlib_encoding')
        pyprofile.incr('vncdriver.recv_rectangle.zlib_encoding.bytes', len(block), unit=pyprofile.BYTES)

        (length,) = struct.unpack('!I', block)
        self.expect(self.recv_DecodeZlib_value, length, x, y, width, height)

    def recv_DecodeZlib_value(self, block, x, y, width, height):
        pyprofile.incr('vncdriver.recv_rectangle.zlib_encoding.bytes', len(block), unit=pyprofile.BYTES)
        rectangle = server_messages.ZlibEncoding.parse_rectangle(self, x, y, width, height, block)
        self._rectangles.append(rectangle)
        self._process_rectangles()

    def recv_SetColorMapEntries(self, block):
        # self.observer.server_data_block(block)

        (first_color, number_of_colors) = struct.unpack('!xHH', block)
        self._handle_SetColorMapEntries(first_color, number_of_colors, [])

    def _handle_SetColorMapEntries(self, first_color, number_of_colors, colors):
        if number_of_colors > 0:
            self.expect(self.recv_SetColorMapEntries_color, 6, first_color, number_of_colors-1, colors)
        else:
            # self.observer.server_set_color_map_entries(first_color, colors)
            self.expect(self.recv_ServerToClient, 1)

    def recv_SetColorMapEntries_color(self, block, first_color, number_of_colors, colors):
        # self.observer.server_data_block(block)

        red, green, blue = struct.unpack('!HHH', block)
        colors.append((red, green, blue))
        self._handle_SetColorMapEntries(first_color, number_of_colors, colors)

    def recv_ServerCutText(self, block):
        # We drop these messages
        (length,) = struct.unpack('!xxxI', block)
        self.expect(self.recv_ServerCutText_value, length)

    def recv_ServerCutText_value(self, block):
        # We drop these messages
        # self.observer.server_cut_text(block)
        self.expect(self.recv_ServerToClient, 1)

    def send_SetEncodings(self, encodings):
        self.sendMessage(struct.pack("!BxH", 2, len(encodings)))
        for encoding in encodings:
            self.sendMessage(struct.pack("!i", encoding))

    def send_PixelFormat(self, bpp=32, depth=24, bigendian=0, truecolor=1, redmax=255, greenmax=255, bluemax=255, redshift=0, greenshift=8, blueshift=16):
        if not self.initialized:
            # Not too bad to add, but no need right now. (We'd need to
            # make sure the framebuffer settings don't get
            # overridden.)
            raise error.Error('Framebuffer not initialized. We have not yet added support for queuing PixelFormat messages before initialization')
        server_pixel_format = struct.pack("!BBBBHHHBBBxxx", bpp, depth, bigendian, truecolor, redmax, greenmax, bluemax, redshift, greenshift, blueshift)
        self.sendMessage(struct.pack("!Bxxx16s", 0, server_pixel_format))
        self.framebuffer.apply_format(server_pixel_format)

    def send_FramebufferUpdateRequest(self, x=0, y=0, width=None, height=None, incremental=0):
        if not self.initialized:
            # Not too bad to add, but no need right now. (We'd need to
            # calculate the message after the framebuffer is
            # initialized.)
            raise error.Error('Framebuffer not initialized. We have not yet added support for queuing FramebufferUpdateRequest messages before initialization')
        if width is None:
            width  = self.framebuffer.width - x
        if height is None:
            height = self.framebuffer.height - y
        self.sendMessage(struct.pack("!BBHHHH", 3, incremental, x, y, width, height))

    def send_KeyEvent(self, key, down):
        """For most ordinary keys, the "keysym" is the same as the
        corresponding ASCII value.  Other common keys are shown in the
        KEY_ constants.
        """
        self.sendMessage(struct.pack('!BBxxI', 4, down, key))

    def send_PointerEvent(self, x, y, buttonmask=0):
        """Indicates either pointer movement or a pointer button press or
           release. The pointer is now at (x-position, y-position),
           and the current state of buttons 1 to 8 are represented by
           bits 0 to 7 of button-mask respectively, 0 meaning up, 1
           meaning down (pressed).
        """
        self.sendMessage(struct.pack('!BBHH', 5, buttonmask, x, y))

    def send_ClientCutText(self, message):
        """The client has new text in its clipboard.
        """
        self.sendMessage(struct.pack("!BxxxI", 6, len(message)))
        self.sendMessage(message)

    def expect(self, type, length, *args, **kwargs):
        logger.debug('Expecting: %s (length=%s)', type.__name__, length)
        assert isinstance(length, int), "Bad length (not an int): {}".format(length)

        self.expected = type
        self.expected_len = length
        self.expected_args = args
        self.expected_kwargs = kwargs

    def close(self):
        self._close = True
        if self.transport:
            self.transport.loseConnection()

    def _error(self, e):
        self.close()
        self.factory.error_buffer.record(e)
        if self.factory.deferred:
            try:
                self.factory.deferred.errback(utils.format_error(e))
            except defer.AlreadyCalledError:
                pass

def client_factory(deferred, error_buffer):
    factory = protocol.ClientFactory()
    factory.deferred = deferred
    factory.error_buffer = error_buffer
    factory.protocol = VNCClient
    return factory



from multiprocessing import pool
from universe import error, rewarder, vectorized

class Joint(vectorized.Wrapper):
    def __init__(self, env_m):
        self.env_m = env_m
        for env in self.env_m:
            if not env._configured:
                raise error.Error('Joint env should have been initialized: {}'.format(env))

        # TODO: generalize this. Doing so requires adding a vectorized
        # space mode.
        self.action_space = env_m[0].action_space
        self.observation_space = env_m[0].observation_space

        self.pool = pool.ThreadPool(min(len(env_m), 5))

    def _close(self):
        if hasattr(self, 'pool'):
            self.pool.close()

    @property
    def spec(self):
        return None

    @spec.setter
    def spec(self, value):
        pass

    def _render(self, mode='human', close=False):
        return self.env_m[0]._render(mode=mode, close=close)

    def _configure(self, **kwargs):
        self.n = sum(env.n for env in self.env_m)
        self.metadata = self.metadata.copy()
        self.metadata['render.modes'] = self.env_m[0].metadata['render.modes']

    def _reset(self):
        # Keep all env[0] action on the main thread, in case we ever
        # need to render. Otherwise we get segfaults from the
        # go-vncdriver.
        reset_m_async = self.pool.map_async(lambda env: env.reset(), self.env_m[1:])
        reset = self.env_m[0].reset()
        reset_m = [reset] + reset_m_async.get()

        observation_n = []
        for observation_m in reset_m:
            observation_n += observation_m
        return observation_n

    def _step(self, action_n):
        observation_n = []
        reward_n = []
        done_n = []
        info_n = []
        info = {}

        action_m = []
        for env in self.env_m:
            action_m.append(action_n[len(action_m):len(action_m)+env.n])

        # Keep all env[0] action on the main thread, in case we ever
        # need to render. Otherwise we get segfaults from the
        # go-vncdriver.
        step_m_async = self.pool.map_async(lambda arg: arg[0].step(arg[1]), zip(self.env_m[1:], action_m[1:]))
        step = self.env_m[0].step(action_m[0])
        step_m = [step] + step_m_async.get()

        for observation_m, reward_m, done_m, _info in step_m:
            observation_n += observation_m
            reward_n += reward_m
            done_n += done_m

            # copy in any info keys
            rewarder.merge_infos(info, _info)
            info_n += _info['n']

        info['n'] = info_n
        return observation_n, reward_n, done_n, info



import errno
import logging
import select
import socket
import time

from universe import error, utils
from gym.utils import reraise

logger = logging.getLogger(__name__)

def run(vnc_addresses, rewarder_addresses, timeout=None, start_timeout=None):
    healthcheck = Healthcheck(vnc_addresses, rewarder_addresses, timeout=timeout, start_timeout=start_timeout)
    healthcheck.run()

def host_port(address, default_port=None):
    split = address.split(':')
    if len(split) == 1:
        host = split[0]
        port = default_port
    else:
        host, port = split
        port = int(port)
    return host, port

class Healthcheck(object):
    def __init__(self, vnc_addresses, rewarder_addresses, timeout=None, start_timeout=None):
        self.timeout = timeout or (4 * len(vnc_addresses) + 20)
        self.start_timeout = start_timeout

        start_time = time.time()

        self.sockets = {}
        for address in vnc_addresses:
            self._register_vnc(address, start_time)
        for address in rewarder_addresses:
            self._register_rewarder(address, start_time)

    def _register_vnc(self, address, start_time=None):
        if start_time is None:
            start_time = time.time()

        host, port = host_port(address, default_port=5900)

        while True:
            # In VNC, the server sends bytes upon connection
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                sock.connect((host, port))
            except (socket.error, socket.gaierror) as e:
                # ECONNREFUSED: VNC env hasn't come up yet
                # ETIMEDOUT: the packets can't be delivered yet, such as can happen on kubernetes
                # gaierror: can't resolve the address yet, which can also happen on kubernetes
                expected = socket.errno.ECONNREFUSED == e.errno or socket.errno.ETIMEDOUT == e.errno or isinstance(e, socket.gaierror)
                if self.start_timeout is None or not expected:
                    reraise(suffix='while connecting to VNC server {}'.format(address))
                logger.info('VNC server %s did not come up yet (error: %s). Sleeping for 1s.', address, e)
                time.sleep(1)
            else:
                break

            if time.time() - start_time > self.start_timeout:
                raise error.Error('VNC server {} did not come up within {}s'.format(address, self.start_timeout))

        self.sockets[sock] = ('vnc', address)

    def _register_rewarder(self, address, start_time=None):
        if start_time is None:
            start_time = time.time()

        host, port = host_port(address, default_port=15900)

        while True:
            # In WebSockets, the server sends bytes once we've upgraded the protocol
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                sock.connect((host, port))
            except (socket.error, socket.gaierror) as e:
                # ECONNREFUSED: VNC env hasn't come up yet
                # ETIMEDOUT: the packets can't be delivered yet, such as can happen on kubernetes
                # gaierror: can't resolve the address yet, which can also happen on kubernetes
                expected = socket.errno.ECONNREFUSED == e.errno or socket.errno.ETIMEDOUT == e.errno or isinstance(e, socket.gaierror)
                if self.start_timeout is None or not expected:
                    reraise(suffix='while connecting to Rewarder server {}'.format(address))
                logger.info('Rewarder server %s did not come up yet (error: %s). Sleeping for 1s.', address, e)
                time.sleep(1)
            else:
                break

            if time.time() - start_time > self.start_timeout:
                raise error.Error('Rewarder server {} did not come up within {}s'.format(address, self.start_timeout))

        # Send a websocket handshake.
        # https://developer.mozilla.org/en-US/docs/Web/API/WebSockets_API/Writing_WebSocket_servers
        #
        # The port 10003 is an arbitrary port that we don't actually connect to, but needs to be a valid part
        # e.g Host: 127.0.0.1:GARBAGE results in the following error: (invalid port 'GARBAGE' in HTTP Host header '127.0.0.1:GARBAGE')
        sock.send(b'GET / HTTP/1.1\r\nHost: 127.0.0.1:10003\r\nUpgrade: WebSocket\r\nConnection:Upgrade\r\nSec-WebSocket-Key: dGhlIHNhbXBsZSBub25jZQ==\r\nSec-WebSocket-Version: 13\r\nauthorization: ' + utils.basic_auth_encode('openai').encode('utf-8') + b'\r\nopenai-observer: true\r\n\r\n')
        self.sockets[sock] = ('rewarder', address)

    def run(self):
        target = time.time() + self.timeout
        while self.sockets:
            remaining = target - time.time()
            if remaining < 0:
                break
            ready, _, _ = select.select(self.sockets.keys(), [], [], remaining)

            # Go through the readable sockets
            remote_closed = False
            for sock in ready:
                type, address = self.sockets.pop(sock)

                # Connection was closed; try again.
                #
                # This is guaranteed not to block.
                try:
                    recv = sock.recv(1)
                except socket.error as e:
                    if e.errno == errno.ECONNRESET:
                        recv = b''
                    else:
                        raise

                if recv == b'':
                    logger.info('Remote closed: address=%s', address)
                    remote_closed = True
                    if type == 'rewarder':
                        self._register_rewarder(address)
                    else:
                        self._register_vnc(address)
                else:
                    logger.debug('Healthcheck passed for %s %s', type, address)

                sock.close()

            if remote_closed:
                sleep = 1
                logger.info('At least one sockets was closed by the remote. Sleeping %ds...', sleep)
                time.sleep(sleep)

        if self.sockets:
            raise error.Error('Not all servers came up within {}s: {}'.format(self.timeout, list(self.sockets.values())))



import logging
import numpy as np

from gym.utils import reraise

from universe import error, rewarder, spaces, utils, vectorized
from universe.envs import diagnostics
from universe.remotes import healthcheck
from universe.runtimes import registration

class DummyVNCEnv(vectorized.Env):
    """
    A simple env for unit testing that does nothing, but looks like a VNC env.
    It accepts any actions, and returns black screens.
    It also returns the actions in the observation, so you can test that action wrappers are producing the right answers
    For example, to test that YourActionWrapper converts example_input_action to example_output_action:

    >>> dummy_env = gym.make('test.DummyVNCEnv-v0')
    >>> e = YourActionWrapper(dummy_env)
    >>> e = universe.wrappers.Unvectorize(e)
    >>> observation, reward, done, info = e.step(example_input_action)
    >>> assert observation['action'] == example_output_action

    """
    metadata = {
        'render.modes': ['human'], # we wrap with a Render which can render to rgb_array
        'semantics.async': True,
        'semantics.autoreset': True,
        'video.frames_per_second' : 60,
        'runtime.vectorized': True,
    }

    def __init__(self):
        self._started = False

        self.observation_space = spaces.VNCObservationSpace()
        self.action_space = spaces.VNCActionSpace()

    def _configure(self, remotes=None,
                   client_id=None,
                   start_timeout=None, docker_image=None,
                   ignore_clock_skew=False, disable_action_probes=False,
                   vnc_driver=None, vnc_kwargs={},
                   replace_on_crash=False, allocate_sync=True,
                   observer=False,
                   _n=3,
    ):
        super(DummyVNCEnv, self)._configure()
        self.n = _n
        self._reward_buffers = [rewarder.RewardBuffer('dummy:{}'.format(i)) for i in range(self.n)]
        self._started = True

    def _reset(self):
        return [None] * self.n

    def _step(self, action_n):
        assert self.n == len(action_n), "Expected {} actions but received {}: {}".format(self.n, len(action_n), action_n)

        observation_n = [{
            'visual': np.zeros((1024, 768, 3), dtype=np.uint8),
            'text': [],
            'action': action_n[i]
        } for i in range(self.n)]

        reward_n = []
        done_n = []
        info_n = []
        for reward_buffer in self._reward_buffers:
            reward, done, info = reward_buffer.pop()
            reward_n.append(reward)
            done_n.append(done)
            info_n.append(info)
        return observation_n, reward_n, done_n, {'n': info_n}

    def __str__(self):
        return 'DummyVNCEnv'



import logging
import six
from universe import pyprofile, vectorized

logger = logging.getLogger(__name__)

# Not used in core; but used in play_flashgames
class Diagnostics(vectorized.Wrapper):
    def _configure(self, **kwargs):
        super(Diagnostics, self)._configure(**kwargs)
        self.diagnostics = self.unwrapped.diagnostics

    def _step(self, action_n):
        observation_n, reward_n, done_n, info = self.env.step(action_n)
        # We want this to be above Mask, so we know whether or not a
        # particular index is resetting.
        if self.diagnostics:
            with pyprofile.push('vnc_env.diagnostics.add_metadata'):
                self.diagnostics.add_metadata(observation_n, info['n'])
        return observation_n, reward_n, done_n, info



import gym
import string

from gym.spaces import prng

from universe.vncdriver import constants
from universe.spaces import vnc_event

class VNCActionSpace(gym.Space):
    """The space of VNC actions.

    You can submit a list of KeyEvents or PointerEvents. KeyEvents
    correspond to pressing or releasing a key. PointerEvents correspond
    to moving to a specific pixel, and setting the mouse buttons to some state
    (buttonmask is a bitmap corresponding to which buttons are down).

    Note that key releases work differently from click releases: keys
    are stateful and must be explicitly released, while the state of
    the mouse buttons is provided at each timestep, so you have to
    explicitly keep the mouse down.

    Attributes:
        keys (list<KeyEvent>): The allowed key presses
        buttonmasks (list<int>): The allowed buttonmasks (i.e. mouse presses)
        screen_shape (int, int): The X and Y dimensions of the screen
    """

    def __init__(self, keys=None, buttonmasks=None, screen_shape=(1024, 728)):
        self.keys = []
        if keys is None:
            keys = [c for c in string.printable] + list(constants.KEYMAP.keys())
        for key in (keys or []):
            down = vnc_event.KeyEvent.by_name(key, down=True)
            up = vnc_event.KeyEvent.by_name(key, down=False)
            self.keys.append(down)
            self.keys.append(up)
        self._key_set = set(self.keys)

        self.screen_shape = screen_shape
        if self.screen_shape is not None:
            self.buttonmasks = []
            if buttonmasks is None:
                buttonmasks = range(256)
            for buttonmask in buttonmasks:
                self.buttonmasks.append(buttonmask)
            self._buttonmask_set = set(self.buttonmasks)

    def contains(self, action):
        if not isinstance(action, list):
            return False

        for a in action:
            if isinstance(a, vnc_event.KeyEvent):
                if a not in self._key_set:
                    return False
            elif isinstance(a, vnc_event.PointerEvent):
                if self.screen_shape is None:
                    return False

                if a.x < 0 or a.x > self.screen_shape[0]:
                    return False
                elif a.y < 0 or a.y > self.screen_shape[1]:
                    return False
                elif a.buttonmask not in self._buttonmask_set:
                    return False

        return True

    def sample(self):
        # Both key and pointer allowed
        if self.screen_shape is not None:
            event_type = prng.np_random.randint(2)
        else:
            event_type = 0

        if event_type == 0:
            # Let's press a key
            key = prng.np_random.choice(self.keys)
            event = [key]
        else:
            x = prng.np_random.randint(self.screen_shape[0])
            y = prng.np_random.randint(self.screen_shape[1])
            buttonmask = prng.np_random.choice(self.buttonmasks)

            event = [vnc_event.PointerEvent(x, y, buttonmask)]
        return event



import logging
import time
import threading

from universe.vncdriver.screen import numpy_screen

logger = logging.getLogger(__name__)

class ScreenBuffer(object):
    def __init__(self):
        self.lock = threading.Lock()
        self.uncommitted = []
        self.updates = []

    def apply_format(self, attrs):
        self._push({
            'type': 'apply_format',
            'attrs': attrs,
        })

    def update_rectangle(self, x, y, width, height, data):
        self._push({
            'type': 'update_rectangle',
            'x': x,
            'y': y,
            'width': width,
            'height': height,
            'data': data,
        })

    def copy_rectangle(self, src_x, src_y, x, y, width, height):
        self._push({
            'type': 'copy_rectangle',
            'src_x': src_x,
            'src_y': src_y,
            'x': x,
            'y': y,
            'width': width,
            'height': height,
        })

    def fill_rectangle(self, x, y, width, height, color):
        self._push({
            'type': 'fill_rectangle',
            'x': x,
            'y': y,
            'width': width,
            'height': height,
            'color': color,
        })

    def framebuffer_update_finish(self):
        with self.lock:
            self.updates += self.uncommitted
            self.uncommitted = []

    def _push(self, update):
        """Always call from single thread."""
        self.uncommitted.append(update)

    def pop(self):
        with self.lock:
            if self.updates:
                updates = self.updates
                self.updates = []
                return updates
            else:
                return None

    def peek(self):
        with self.lock:
            if self.updates:
                return self.updates
            else:
                return None



import gym
from gym.spaces import Box
from universe.spaces import joystick_event
from gym.spaces import prng
from collections import OrderedDict


class JoystickActionSpace(gym.Space):
    """
    Programmable joystick - currently Windows-only => mapped to vJoy
    """
    def __init__(self, axis_x=False, axis_y=False, axis_z=False, axis_rx=False, axis_ry=False, axis_rz=False,
                 slider_0=False, slider_1=False):
        self.event_space_map = OrderedDict()

        if axis_x:
            self.axis_x = box_axis()
            self.event_space_map[joystick_event.JoystickAxisXEvent] = self.axis_x
        if axis_y:
            self.axis_y = box_axis()
            self.event_space_map[joystick_event.JoystickAxisYEvent] = self.axis_y
        if axis_z:
            self.axis_z = box_axis()
            self.event_space_map[joystick_event.JoystickAxisZEvent] = self.axis_z
        if axis_rx:
            self.axis_rx = box_axis()
            self.event_space_map[joystick_event.JoystickAxisRxEvent] = self.axis_rx
        if axis_ry:
            self.axis_ry = box_axis()
            self.event_space_map[joystick_event.JoystickAxisRyEvent] = self.axis_ry
        if axis_rz:
            self.axis_rz = box_axis()
            self.event_space_map[joystick_event.JoystickAxisRzEvent] = self.axis_rz
        if slider_0:
            self.slider_0 = box_axis()
            self.event_space_map[joystick_event.JoystickSlider0Event] = self.slider_0
        if slider_1:
            self.slider_1 = box_axis()
            self.event_space_map[joystick_event.JoystickSlider1Event] = self.slider_1
        # TODO: Add buttons (similar to a vnc_event.KeyEvent - but 1..32)
        # TODO: Add POV hats

    def contains(self, action):
        if not isinstance(action, list):
            return False
        for a in action:
            if isinstance(a, joystick_event.JoystickAxisEvent):
                axis = self.event_space_map[a]
                if not axis.contains(a):
                    return False
        return True

    def sample(self):
        event_type_index = prng.np_random.randint(len(self.event_space_map))
        event_type = list(self.event_space_map.keys())[event_type_index]
        if event_type.__bases__[0] == joystick_event.JoystickAxisEvent:
            event = [event_type(self.event_space_map[event_type].sample()[0])]
        else:
            raise JoystickActionSpaceException('Unexpected event type')
        return event


class JoystickActionSpaceException(Exception):
    pass


def box_axis():
    return Box(-1.0, 1.0, shape=(1,))



import logging
import time

import gym
from universe import spaces
from universe.envs import vnc_env

logger = logging.getLogger(__name__)

class GymCoreEnv(vnc_env.VNCEnv):
    def __init__(self, gym_core_id, fps=60):
        super(GymCoreEnv, self).__init__()

        self.metadata = dict(self.metadata)
        self.metadata['video.frames_per_second'] = fps

        self.gym_core_id = gym_core_id
        self._seed_value = None

        self.vnc_pixels = True

class GymCoreSyncEnv(GymCoreEnv):
    """A synchronized version of the core envs. Its semantics should match
    that of the core envs. (By default, observations are pixels from
    the VNC session, but it also supports receiving the normal Gym
    observations over the rewarder socket.)

    Provided primarily for testing and debugging.
    """

    def __init__(self, gym_core_id, fps=60, vnc_pixels=True):
        super(GymCoreSyncEnv, self).__init__(gym_core_id, fps=fps)
        # Metadata has already been cloned
        self.metadata['semantics.async'] = False

        self.gym_core_id = gym_core_id
        self.vnc_pixels = vnc_pixels

        if not vnc_pixels:
            self._core_env = gym.spec(gym_core_id).make()
        else:
            self._core_env = None

    def _flip_past(self, when_n):
        info_n = [{} for i in range(self.n)]
        while True:
            observation_n, obs_info_n = self.vnc_session.flip()
            metadata_n = self.diagnostics.extract_metadata(observation_n)

            # Save the update count
            self._propagate_obs_info(info_n, obs_info_n)

            # All remote times, so no clock skew adjustments needed
            invalid = []
            for i, (metadata, when) in enumerate(zip(metadata_n, when_n)):
                delta = when - metadata.get('now', 0)
                if delta > 0:
                    invalid.append((i, delta))
            if not invalid:
                break
            else:
                tick = 1./self.metadata['video.frames_per_second']
                logger.info('Waiting %sms for the following observations to catch up: %s', int(1000*tick), invalid)
                time.sleep(tick)
        return observation_n, info_n

    def _reset(self):
        assert self.rewarder_session

        result = self.rewarder_session.reset(
            seed=self._seed_value,
        )
        # Clear seed value so we don't double-send it
        self._seed_value = None

        # Wait until all the observations have passed the reset_time
        remote_reset_time = [response['headers']['sent_at'] for _, _, response in result]
        observation_n, _ = self._flip_past(remote_reset_time)

        # Double check that our reward queue is empty
        assert all(c == 0 for c in self.rewarder_session.rewards_count())

        return self._observation(observation_n)

    def _observation(self, observation_n):
        if self.vnc_pixels:
            return observation_n
        else:
            observation_n = self.rewarder_session.pop_observation()
            assert all(observation is not None for observation in observation_n), 'At least one missing observation: {}'.format(observation_n)
            return self._core_env.observation_space.from_jsonable(observation_n)

    def _step(self, action_n):
        # Add C keypress in order to "commit" the action, as
        # interpreted by the remote.
        action_n = [action + [
            spaces.KeyEvent.by_name('c', down=True),
            spaces.KeyEvent.by_name('c', down=False)
        ] for action in action_n]
        # Submit directly to VNC session, without popping rewards
        logger.debug('Submitting actions: %s', action_n)
        action_n = self._compile_actions(action_n)
        _, obs_info_n = self.vnc_session.step(action_n)
        # Wait until the actions have actually happened
        self.rewarder_session.wait(timeout=5)

        # TODO: this is now present in the info messages; need to
        # update the implementation.
        when_n = [reward_buffer.info['reward_buffer.remote_time'] for reward_buffer in self.rewarder_session.reward_buffers]
        observation_n, obs_info_n = self._flip_past(when_n)

        reward_n, reward_time_n, done_n, info_n = self.rewarder_session.pop()
        self._propagate_obs_info(info_n, obs_info_n)

        # Warn if we detect multiple rewards
        if any(info['stats.reward.count'] != 1 for info in info_n):
            # Arrived but there was a bug
            logger.warn('Likely bug: should have received 1 reward for every env, but instead received %s. Current return: observation=%s reward=%s done=%s info=%s', [info['stats.reward.count'] for info in info_n], observation_n, reward_n, done_n, info_n)

        return self._observation(observation_n), reward_n, done_n, {'n': info_n}



import logging
import os
import re
import shutil
import struct
import time
import traceback

from universe import pyprofile
from universe.vncdriver import auth, constants, fbs_writer
from twisted.internet import defer, endpoints, protocol, reactor

logger = logging.getLogger(__name__)

class LogManager(object):
    def __init__(self, base_logfile_dir, recorder_id, id):
        self.base_logfile_dir = base_logfile_dir
        self.recorder_id = recorder_id
        self.id = id

        self.logfile_dir = os.path.join(self.base_logfile_dir, '{}-{}-{}'.format(int(time.time()), self.recorder_id, str(self.id)))

        logger.info('[vnc_proxy] logfile_dir = %s', self.logfile_dir)
        # Make the logfile directory that will be written to
        if not os.path.isdir(self.logfile_dir):
            logger.info('[vnc_proxy] [%s] Creating log directory %s', self.id, self.logfile_dir)
            os.makedirs(self.logfile_dir)

    @property
    def global_rewards_logfile(self):
        """
        We assume that there is only a single rewards file that is ever written to.
        This is enforced by the reward_recorder that limits connections to one
        """
        return '/tmp/demo/rewards.demo'

    @property
    def global_botaction_logfile(self):
        """
        We assume that there is only a single botactions file that is ever written to.
        This is enforced by the botaction_recorder that appends everything to one file
        """
        return '/tmp/demo/botactions.jsonl'


    @property
    def global_env_id_file(self):
        """
        We assume that there is only a single rewards file that is ever written to.
        This is enforced by the reward_recorder that limits connections to one
        """
        return '/tmp/demo/env_id.txt'

    @property
    def server_logfile(self):
        return os.path.join(self.logfile_dir, 'server.fbs')

    @property
    def client_logfile(self):
        return os.path.join(self.logfile_dir, 'client.fbs')

    def close(self):
        logger.info("[%s] Copying rewards.demo into this connection's log dir", self.id)

        if os.path.exists(self.global_rewards_logfile):
            # Each recording includes all the rewards that we've seen. We count on the player to
            # crop them to only contain the ones that occurred during this recording

            # TODO: Upload to S3 immediately upon disconnect
            shutil.copyfile(self.global_rewards_logfile, os.path.join(self.logfile_dir, 'rewards.demo'))
        else:
            logger.info("%s does not exist; not copying into recording directory", self.global_rewards_logfile)

        if os.path.exists(self.global_botaction_logfile):
            shutil.copyfile(self.global_botaction_logfile, os.path.join(self.logfile_dir, 'botactions.jsonl'))
        else:
            logger.info("%s does not exist; not copying into recording directory", self.global_botaction_logfile)


        if os.path.exists(self.global_env_id_file):
            shutil.copyfile(self.global_env_id_file, os.path.join(self.logfile_dir, 'env_id.txt'))
        else:
            logger.info("%s does not exist; not copying into recording directory", self.global_env_id_file)

        if os.environ.get('COMPLETED_DEMONSTRATION_DIR', None):
            shutil.copytree(self.logfile_dir, os.path.join(os.environ['COMPLETED_DEMONSTRATION_DIR'], os.path.basename(self.logfile_dir)))


class VNCProxyServer(protocol.Protocol, object):
    """Bytes received from the end user. (So received data are mostly
    actions.)"""

    _next_id = 0

    SUPPORTED_ENCODINGS = {
        # Maybe we can do copy-rect at some point. May not help with
        # much though.
        # constants.COPY_RECTANGLE_ENCODING,
        constants.TIGHT_ENCODING,
        constants.RAW_ENCODING,
        constants.ZLIB_ENCODING,
        # constants.HEXTILE_ENCODING,
        # constants.CORRE_ENCODING,
        # constants.RRE_ENCODING,
        constants.PSEUDO_CURSOR_ENCODING
    }

    if os.getenv('VNC_ENCODINGS'):
        for enc in os.getenv('VNC_ENCODINGS').split():
            SUPPORTED_ENCODINGS.add(getattr(constants, enc.upper() + '_ENCODING'))
    add_pseudo_cursor_encoding = True
    if os.getenv('VNC_ENCODINGS_NO_ADD_PSEUDO_CURSOR_ENCODING'):
        add_pseudo_cursor_encoding = False

    # ZRLE format is much smaller, but slower to read.
    # SUPPORTED_ENCODINGS.add(constants.ZRLE_ENCODING)

    @classmethod
    def next_id(cls):
        id = cls._next_id
        cls._next_id += 1
        return id

    def __init__(self, action_queue=None, error_buffer=None, enable_logging=True):
        self.id = self.next_id()

        self.server_log = None
        self.action_queue = action_queue
        self.error_buffer = error_buffer
        self.server_log_buffer = []
        self.enable_logging = enable_logging

        self._broken = False
        self.vnc_client = None
        self.log_manager = None

        self.buf = []
        self.buf_len = 0

        self.queued_data = []
        self.initialized = False

        self.challenge = auth.challenge()
        self.accept_any_password = True
        self.expect(self.recv_ProtocolVersion_Handshake, 12)

    def start_logging(self):
        self.log_manager = LogManager(self.factory.logfile_dir, self.factory.recorder_id, self.id)
        self.server_log = fbs_writer.FBSWriter(self.log_manager.server_logfile)

        for data in self.server_log_buffer:
            self.server_log.write(data)
        # Done with the buffer!
        self.server_log_buffer = None


    def connectionMade(self):
        logger.info('[%s] Connection received from VNC client', self.id)
        factory = protocol.ClientFactory()
        factory.protocol = VNCProxyClient
        factory.vnc_server = self
        factory.deferrable = defer.Deferred()
        endpoint = endpoints.clientFromString(reactor, self.factory.vnc_address)

        def _established_callback(client):
            if self._broken:
                client.close()
            self.vnc_client = client
            self.flush()
        def _established_errback(reason):
            logger.error('[VNCProxyServer] Connection succeeded but could not establish session: %s', reason)
            self.close()
        factory.deferrable.addCallbacks(_established_callback, _established_errback)

        def _connect_errback(reason):
            logger.error('[VNCProxyServer] Connection failed: %s', reason)
            self.close()
        endpoint.connect(factory).addErrback(_connect_errback)

        self.send_ProtocolVersion_Handshake()

    def connectionLost(self, reason):
        logger.info('Losing connection from VNC user')
        if self.vnc_client:
            self.vnc_client.close()

    def dataReceived(self, data):
        pyprofile.incr('vnc_proxy_server.data.sent.messages')
        pyprofile.incr('vnc_proxy_server.data.sent.bytes', len(data))

        self.buf.append(data)
        self.buf_len += len(data)
        self.flush()

    def sendData(self, data):
        if self.server_log is None:
            self.server_log_buffer.append(data)
        else:
            self.server_log.write(data)
        self.transport.write(data)

    def flush(self):
        if self.buf_len < self.expected_len:
            return
        elif self.vnc_client is None and self.action_queue is None:
            return

        buffer = b''.join(self.buf)
        while not self._broken and len(buffer) >= self.expected_len:
            block, buffer = buffer[:self.expected_len], buffer[self.expected_len:]
            self.handle(self.expected, block)

        self.buf[:] = [buffer]
        self.buf_len = len(buffer)

    def handle(self, type, block):
        logger.debug('[%s] Handling: type=%s', self.id, type)
        try:
            self.expected(block, *self.expected_args, **self.expected_kwargs)
        except Exception as e:
            self._error(e)

    def send_ProtocolVersion_Handshake(self):
        self.sendData(b'RFB 003.003\n')

    def recv_ProtocolVersion_Handshake(self, block):
        # Client chooses RFB version
        match = re.search(b'^RFB (\d{3}).(\d{3})\n$', block)
        assert match, 'Block does not match: {!r}'.format(block)
        major = int(match.group(1))
        minor = int(match.group(2))
        self.protocol_version = (major, minor)
        assert major == 3 and minor in (3, 8), 'Unexpected version: {}'.format((major, minor))

        if minor == 3:
            self.send_VNC_Authentication()
        elif minor == 8:
            self.send_SecurityTypes()

    def send_SecurityTypes(self):
        self.sendData(struct.pack('!BB', 1, 2))
        self.expect(self.recv_SecurityTypesResponse, 1)

    def recv_SecurityTypesResponse(self, block):
        (type,) = struct.unpack('!B', block)
        assert type == 2
        self.send_VNC_Authentication()

    def send_VNC_Authentication(self):
        # Now we tell the client to auth with password. (Only do this
        # because the built-in Mac viewer prompts for a password no
        # matter what.)
        self.sendData(struct.pack('!I', 2))
        self.sendData(self.challenge)
        self.expect(self.recv_VNC_Authentication_response, 16)

    def recv_VNC_Authentication_response(self, block):
        expected = auth.challenge_response(self.challenge)
        if block == expected or self.accept_any_password:
            logger.debug('Client authenticated successfully')
            self.send_SecurityResult_Handshake_success()
        else:
            logger.debug('VNC client supplied incorrect password')
            self.send_SecurityResult_Handshake_failed('Your password was incorrect')

    def send_SecurityResult_Handshake_success(self):
        logger.debug('[%s] Send SecurityResult_Handshake_success', self.id)
        self.sendData(struct.pack('!I', 0))
        self.expect(self.recv_ClientInit, 1)

    def send_SecurityResult_Handshake_failed(self, reason):
        self.sendData(struct.pack('!I', 1))
        self.sendData(struct.pack('!I', len(reason)))
        self.sendData(reason)
        self.close()

    def recv_ClientInit(self, block):
        (shared,) = struct.unpack('!B', block)

        ### Now that the server is up and running, we flush
        ### everything.

        # We don't even create log directory until we've successfully
        # established the connection.

        if self.enable_logging and self.factory.logfile_dir:
            # Log in vnc_recorder; don't log in playback
            self.start_logging()
            self.vnc_client.start_logging()

        for data in self.queued_data:
            self.sendData(data)
        self.queued_data = None
        self.initialized = True

        # Listen for messages
        self.expect(self.recv_ClientToServer, 1)

    def recv_ClientToServer(self, block):
        (message_type,) = struct.unpack('!B', block)
        if message_type == 0:
            self.proxyData(block)
            self.expect(self.recv_SetPixelFormat, 19)
        elif message_type == 2:
            # Do not proxy since we want to transform it
            self.expect(self.recv_SetEncodings, 3)
        elif message_type == 3:
            self.proxyData(block)
            self.expect(self.recv_FramebufferUpdateRequest, 9)
        elif message_type == 4:
            self.proxyData(block)
            self.expect(self.recv_KeyEvent, 7)
        elif message_type == 5:
            self.proxyData(block)
            self.expect(self.recv_PointerEvent, 5)
        elif message_type == 6:
            # Do not proxy since we don't support it
            self.expect(self.recv_ClientCutText, 7)
        else:
            assert False, 'Unknown client to server message type received: {}'.format(message_type)

    def recv_SetPixelFormat(self, block):
        self.proxyData(block)

        (server_pixel_format,) = struct.unpack('!xxx16s', block)

        if self.action_queue:
            self.action_queue.set_pixel_format(server_pixel_format)

        # self.vnc_client.framebuffer.apply_format(server_pixel_format)
        self.expect(self.recv_ClientToServer, 1)

    def recv_SetEncodings(self, block):
        # Do not proxy, as we will write our own transformed version
        # of this shortly.

        (number_of_encodings,) = struct.unpack('!xH', block)
        self._handle_SetEncodings(number_of_encodings, [])

    def _handle_SetEncodings(self, number_of_encodings, encodings):
        if number_of_encodings > 0:
            self.expect(self.recv_SetEncodings_encoding_type, 4, number_of_encodings-1, encodings)
        else:
            supported = []
            unsupported = []
            for encoding in encodings:
                if encoding in self.SUPPORTED_ENCODINGS:
                    supported.append(encoding)
                else:
                    unsupported.append(encoding)

            if unsupported:
                logger.info('[%s] Requested %s unsupported encodings: unsupported=%s supported=%s', self.id, len(unsupported), unsupported, supported)

            if self.add_pseudo_cursor_encoding and constants.PSEUDO_CURSOR_ENCODING not in supported:
                logger.info('[%s] Add PSEUDO_CURSOR_ENCODING', self.id)
                supported.append(constants.PSEUDO_CURSOR_ENCODING)

            logger.debug('[%s] Encodings: %s', self.id, supported)
            if self.vnc_client:
                self.vnc_client.send_SetEncodings(supported)
            self.expect(self.recv_ClientToServer, 1)

    def recv_SetEncodings_encoding_type(self, block, number_of_encodings, encodings):
        # Do not proxy, as we will write our own transformed version
        # of this shortly.

        (encoding,) = struct.unpack('!i', block)
        encodings.append(encoding)
        self._handle_SetEncodings(number_of_encodings, encodings)

    def recv_FramebufferUpdateRequest(self, block):
        self.proxyData(block)

        incremental, x, y, width, height = struct.unpack('!BHHHH', block)
        self.expect(self.recv_ClientToServer, 1)

    def recv_KeyEvent(self, block):
        self.proxyData(block)

        down, key = struct.unpack('!BxxI', block)
        if self.action_queue is not None:
            self.action_queue.key_event(key, down)
        self.expect(self.recv_ClientToServer, 1)

    def recv_PointerEvent(self, block):
        self.proxyData(block)

        buttonmask, x, y = struct.unpack('!BHH', block)
        if self.action_queue is not None:
            self.action_queue.pointer_event(x, y, buttonmask)
        self.expect(self.recv_ClientToServer, 1)

    def recv_ClientCutText(self, block):
        # Drop ClientCutText

        (length,) = struct.unpack('!xxxI', block)
        self.expect(self.recv_ClientCutText_value, length)

    def recv_ClientCutText_value(self, block):
        # Drop ClientCutText

        self.expect(self.recv_ClientToServer, 1)

    def expect(self, type, length, *args, **kwargs):
        self.expected = type
        self.expected_len = length
        self.expected_args = args
        self.expected_kwargs = kwargs

    def proxyData(self, data):
        if self.vnc_client:
            self.vnc_client.recvProxyData(data)

    def recvProxyData(self, data):
        """Write data to server"""
        if self.initialized:
            self.sendData(data)
        else:
            self.queued_data.append(data)

    def close(self):
        logger.info('[%s] Closing', self.id)
        self._broken = True
        if self.transport:
            self.transport.loseConnection()
        if self.log_manager:
            self.log_manager.close()
        if self.server_log:
            self.server_log.close()

    def _error(self, e):
        logger.error('[%s] Connection from client aborting with error: %s', self.id, e)
        traceback.print_exc()
        if self.error_buffer:
            self.error_buffer.record(e)
        self.close()

class VNCProxyClient(protocol.Protocol, object):
    def __init__(self):
        self.id = None
        self.vnc_server = None
        self.client_log = None
        # Messages heading for the server, which haven't been written
        # to disk.
        self.client_log_buffer = []
        self.buf = []
        self.buf_len = 0
        self._broken = False

        self.queued_data = []
        self.initialized = False

        self.expect(self.recv_ProtocolVersion_Handshake, 12)

    def start_logging(self):
        self.client_log = fbs_writer.FBSWriter(self.vnc_server.log_manager.client_logfile)
        for data in self.client_log_buffer:
            self.client_log.write(data)
        # Done with the buffer!
        self.client_log_buffer = None

    def connectionMade(self):
        logger.debug('Connection made to VNC server')
        self.vnc_server = self.factory.vnc_server
        self.id = '{}-client'.format(self.vnc_server.id)

    def connectionLost(self, reason):
        logger.info('Losing connection to VNC server')
        if self.vnc_server:
            self.vnc_server.close()

    def proxyData(self, data):
        """Write data to client"""
        assert self.initialized
        self.vnc_server.recvProxyData(data)

    def recvProxyData(self, data):
        """Write data to client"""
        self.sendData(data)

    def sendData(self, data):
        """Write data to server"""
        # Not set up yet
        if self.client_log is None:
            self.client_log_buffer.append(data)
        else:
            self.client_log.write(data)
        self.transport.write(data)

    def dataReceived(self, data):
        if self.expected is None:
            # We're in direct proxy mode
            self.proxyData(data)
            return

        pyprofile.incr('vnc_proxy_server.data.sent.messages')
        pyprofile.incr('vnc_proxy_server.data.sent.bytes', len(data))

        self.buf.append(data)
        self.buf_len += len(data)
        self.flush()

    def flush(self):
        if self.buf_len < self.expected_len:
            return

        buffer = b''.join(self.buf)
        while self.expected is not None and not self._broken and len(buffer) >= self.expected_len:
            block, buffer = buffer[:self.expected_len], buffer[self.expected_len:]
            self.handle(self.expected, block)

        self.buf[:] = [buffer]
        self.buf_len = len(buffer)

        if self.expected is None:
            data = b''.join(self.buf)
            if data != '':
                self.proxyData(data)
            self.buf = []

    def handle(self, type, block):
        logger.debug('[%s] Handling: type=%s', self.id, type)
        try:
            self.expected(block, *self.expected_args, **self.expected_kwargs)
        except Exception as e:
            self._error(e)

    def recv_ProtocolVersion_Handshake(self, block):
        match = re.search(b'^RFB (\d{3}).(\d{3})\n$', block)
        assert match, 'Expected RFB line, but got: {!r}'.format(block)
        major = int(match.group(1))
        minor = int(match.group(2))
        self.sendData(b'RFB 003.003\n')

        self.expect(self.recv_Security_Handshake, 4)

    def recv_Security_Handshake(self, block):
        (auth,) = struct.unpack('!I', block)
        if auth == 0:
            self.expect(self.recv_SecurityResult_Handshake_failed_length, 4)
        elif auth == 1:
            self.send_ClientInit()
        elif auth == 2:
            self.expect(self.recv_VNC_Authentication, 16)
        else:
            assert False, 'Bad auth: {}'.format(auth)

    def recv_VNC_Authentication(self, block):
        response = auth.challenge_response(block)
        self.sendData(response)
        self.expect(self.recv_SecurityResult_Handshake, 4)

    def recv_SecurityResult_Handshake(self, block):
        (result,) = struct.unpack('!xxxB', block)
        if result == 0:
            logger.debug('VNC Auth succeeded')
            self.send_ClientInit()
        elif result == 1:
            logger.debug('VNC Auth failed.')
            # Server optionally can say why
            self.expect(self.recv_SecurityResult_Handshake_failed_length, 4)
        else:
            assert False, 'Bad security result: {}'.format(result)

    def recv_SecurityResult_Handshake_failed_length(self, block):
        (length,) = struct.unpack('!I', block)
        self.expect(self.recv_SecurityResult_Handshake_failed_reason, length)

    def recv_SecurityResult_Handshake_failed_reason(self, block):
        logger.info('Connection to server failed: %s', block)

    def send_ClientInit(self):
        shared = True
        self.sendData(struct.pack('!B', shared))

        ### Now that the session is up and running, we flush
        ### everything and stop parsing.

        # We're up and running!
        logger.info('[%s] Marking server as connected', self.id)
        self.factory.deferrable.callback(self)

        self.initialized = True
        # Flush queue
        for data in self.queued_data:
            self.sendData(data)
        self.queued_data = None

        # And from now on just do a straight proxy
        self.expect(None, None)

    def expect(self, type, length, *args, **kwargs):
        if type is not None:
            logger.debug('Expecting: %s (length=%s)', type.__name__, length)
            assert isinstance(length, int), "Bad length: {}".format(length)

        self.expected = type
        self.expected_len = length
        self.expected_args = args
        self.expected_kwargs = kwargs

    def close(self):
        logger.info('[%s] Closing', self.id)
        self._broken = True
        if self.transport:
            self.transport.loseConnection()
        if self.client_log:
            self.client_log.close()

    def _error(self, e):
        logger.error('[%s] Connection to server aborting with error: %s', self.id, e)
        traceback.print_exc()
        self.close()

    def send_SetEncodings(self, encodings):
        self.sendData(struct.pack("!BxH", 2, len(encodings)))
        for encoding in encodings:
            self.sendData(struct.pack("!i", encoding))



# A simple in-memory stats library
#
# Inspired by statsd: http://statsd.readthedocs.io/en/v3.1/types.html#gauges
import collections
import json
import logging
import numbers
import numpy as np
import os
import six
import threading
import time

logger = logging.getLogger(__name__)

BYTES = 'bytes'
SECONDS = 'seconds'

class Error(Exception):
    pass

class ExponentialAverage(object):
    def __init__(self, decay=0.1):
        self.decay = decay
        self.last_update = None
        self.last_data_decay = None
        self._avg = None

    def add(self, data):
        assert isinstance(data, numbers.Number)
        if self.last_update is None:
            self._avg = data
            self.last_update = time.time()
            self.last_data_decay = 1
        else:
            now = time.time()
            delta = now - self.last_update
            if delta < 0:
                # Time is allowed to go a little backwards (NTP update, etc)
                logger.warn("Backwards delta value: {}".format(delta))
                # Treat this entry as if it happened with 0 delta
                delta = 0
            if delta != 0:
                self.last_data_decay = (1 - self.decay**delta) * 1/delta
                self._avg = self.decay**delta * self._avg + self.last_data_decay * data
            else:
                # Don't divide by zero; just reuse the last delta. Should stack well
                self._avg += self.last_data_decay * data
            self.last_update = now

    def avg(self):
        return self._avg

class RunningVariance(object):
    """ Implements Welford's algorithm for computing a running mean
    and standard deviation as described at:
        http://www.johndcook.com/standard_deviation.html
    can take single values or iterables
    Properties:
        mean    - returns the mean
        std     - returns the std
        meanfull- returns the mean and std of the mean
    Usage:
        >>> foo = Welford()
        >>> foo(range(100))
        >>> foo
        <Welford: 49.5 +- 29.0114919759>
        >>> foo([1]*1000)
        >>> foo
        <Welford: 5.40909090909 +- 16.4437417146>
        >>> foo.mean
        5.409090909090906
        >>> foo.std
        16.44374171455467
        >>> foo.meanfull
        (5.409090909090906, 0.4957974674244838)
    """

    def __init__(self):
        self.k = 0
        self.M = 0
        self.S = 0

    def add(self,x):
        if x is None:
            return
        self.k += 1
        newM = self.M + (x - self.M)*1./self.k
        newS = self.S + (x - self.M)*(x - newM)
        self.M, self.S = newM, newS

    def mean(self):
        return self.M
    def meanfull(self):
        return self.mean, self.std/np.sqrt(self.k)
    def std(self):
        if self.k==1:
            return 0
        return np.sqrt(self.S/(self.k-1))
    def __repr__(self):
        return "<Welford: {} +- {}>".format(self.mean, self.std)

def pretty(d, unit):
    if unit is None:
        return d
    elif unit == BYTES:
        return pretty_bytes(d)
    elif unit == SECONDS:
        return pretty_seconds(d)
    else:
        raise Error('No such unit: {}'.format(unit))

def pretty_bytes(b):
    if b is None:
        return None

    assert isinstance(b, numbers.Number), "Surprising type for data: {} ({!r})".format(type(b), b)
    if b > 1000 * 1000:
        return '{:.0f}MB'.format(b/1000.0/1000.0)
    elif b > 1000:
        return '{:.0f}kB'.format(b/1000.0)
    else:
        return '{:.0f}B'.format(b)

def pretty_seconds(t):
    a_t = abs(t)
    if a_t < 0.001:
        return '{:.2f}us'.format(1000*1000*t)
    elif a_t < 1:
        return '{:.2f}ms'.format(1000*t)
    else:
        return '{:.2f}s'.format(t)

def thread_id():
    return threading.current_thread().ident

class StackProfile(object):
    def __init__(self, profile):
        self.profile = profile

        self.stack_by_thread = {}
        self.lock = threading.Lock()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.pop()

    def push(self, event):
        stack = self._current_stack()
        stack.append({
            'name': event,
            'start': time.time(),
        })
        return self

    def pop(self):
        stack = self._current_stack()
        event = stack.pop()
        name = event['name']
        start = event['start']

        with self.profile as txn:
            delta = time.time() - start
            txn.timing(name, delta)
            # These are now subsumed by the timers key
            # txn.incr(name + '.total_time', delta, unit=SECONDS)
            # txn.incr(name + '.calls')

    def _current_stack(self):
        id = thread_id()

        try:
            stack = self.stack_by_thread[id]
        except KeyError:
            with self.lock:
                # Only current thread should be adding to this entry anyway
                assert id not in self.stack_by_thread
                stack = self.stack_by_thread[id] = []
        return stack

class Profile(object):
    def __init__(self, print_frequency=None, print_filter=None):
        if print_filter is None:
            print_filter = lambda event: True

        self.lock = threading.RLock()

        self.print_frequency = print_frequency
        self.last_export = None

        self.print_filter = print_filter
        self._in_txn = False

        self.reset()

    def reset(self):
        self.timers = {}
        self.counters = {}
        self.gauges = {}

    def __enter__(self):
        self.lock.acquire()
        self._in_txn = True
        return self

    def __exit__(self, type, value, tb):
        self._in_txn = False
        self._print_if_needed()
        self.lock.release()

    def timing(self, event, time):
        assert isinstance(event, six.string_types)
        # return
        with self.lock:
            if event not in self.timers:
                self.timers[event] = {
                    'total': 0,
                    'calls': 0,
                    'std': RunningVariance(),
                }
            self.timers[event]['total'] += time
            self.timers[event]['calls'] += 1
            self.timers[event]['std'].add(time)

            self._print_if_needed()

    def incr(self, event, amount=1, unit=None):
        assert isinstance(event, six.string_types)
        # return
        with self.lock:
            if event not in self.counters:
                self.counters[event] = {
                    'total': 0,
                    'calls': 0,
                    'rate': ExponentialAverage(),
                    'unit': unit,
                    'std': RunningVariance(),
                }
            self.counters[event]['total'] += amount
            self.counters[event]['calls'] += 1
            self.counters[event]['rate'].add(amount)
            self.counters[event]['std'].add(amount)

            self._print_if_needed()

    def gauge(self, event, value, delta=False, unit=None):
        assert isinstance(event, six.string_types)
        with self.lock:
            if event not in self.gauges:
                self.gauges[event] = {
                    'value': 0,
                    'calls': 0,
                    'unit': unit,
                    'std': RunningVariance(),
                }
            if delta:
                self.gauges[event]['value'] += value
            else:
                self.gauges[event]['value'] = value
            self.gauges[event]['calls'] += 1
            self.gauges[event]['std'].add(value)

            self._print_if_needed()

    def _print_if_needed(self):
        """Assumes you hold the lock"""
        if self._in_txn or self.print_frequency is None:
            return
        elif self.last_export is not None and \
             self.last_export + self.print_frequency > time.time():
            return

        self.export()

    def export(self, log=True, reset=True):
        with self.lock:
            if self.last_export is None:
                self.last_export = time.time()
            delta = time.time() - self.last_export
            self.last_export = time.time()

            timers = {}
            for event, stat in self.timers.items():
                timers[event] = {
                    'mean': stat['std'].mean(),
                    'std': stat['std'].std(),
                    'calls': stat['calls'],
                    'unit': 'seconds',
                }

            counters = {}
            for counter, stat in self.counters.items():
                counters[counter] = {
                    'calls': stat['calls'],
                    'std': stat['std'].std(),
                    'mean': stat['std'].mean(),
                    'unit': stat['unit'],
                }

            gauges = {}
            for gauge, stat in self.gauges.items():
                gauges[gauge] = {
                    'value': stat['value'],
                    'calls': stat['calls'],
                    'std': stat['std'].std(),
                    'mean': stat['std'].mean(),
                    'unit': stat['unit'],
                }

            export = {
                'timers': timers,
                'counters': counters,
                'gauges': gauges,
                'metadata': {
                    'period': delta,
                }
            }
            if log:
                self._print_export(export)
            if reset:
                self.reset()
            return export

    def _print_export(self, export):
        timers = {}
        for event, stat in sorted(export['timers'].items()):
            if not self.print_filter(event):
                continue

            timers[event] = {
                'mean': pretty_seconds(stat['mean']),
                'std': pretty_seconds(stat['std']),
                'calls': stat['calls'],
            }

        counters = collections.OrderedDict({})
        for counter, stat in sorted(export['counters'].items()):
            if not self.print_filter(counter):
                continue

            unit = stat['unit']
            counters[counter] = {
                'calls': stat['calls'],
                'std': pretty(stat['std'], unit),
                'mean': pretty(stat['mean'], unit),
            }

        gauges = collections.OrderedDict({})
        for gauge, stat in sorted(export['gauges'].items()):
            if not self.print_filter(gauge):
                continue

            unit = stat['unit']
            gauges[gauge] = {
                'value': pretty(stat['value'], unit),
                'calls': stat['calls'],
                'std': pretty(stat['std'], unit),
                'mean': pretty(stat['mean'], unit),
            }

        # A bit of a hack, but we want this time to be as inclusive as
        # possible.
        export['metadata']['export_time'] = time.time() - self.last_export

        # We do the explicit OrderedDict and json.dumps to order
        # keys. Maybe there's a better way?
        logger.info('[pyprofile] period=%s timers=%s counters=%s gauges=%s (export_time=%s)',
                    pretty_seconds(export['metadata']['period']),
                    json.dumps(timers), json.dumps(counters), json.dumps(gauges),
                    pretty_seconds(export['metadata']['export_time']),
        )

print_frequency = os.environ.get('PYPROFILE_FREQUENCY')
if print_frequency is not None:
    print_frequency = int(print_frequency)

print_prefix = os.environ.get('PYPROFILE_PREFIX')
if print_prefix is not None:
    print_filter = lambda event: event.startswith(print_prefix)
else:
    print_filter = None

profile = Profile(print_frequency=print_frequency, print_filter=print_filter)
stack_profile = StackProfile(profile)

push = stack_profile.push
pop = stack_profile.pop
incr = profile.incr
timing = profile.timing
gauge = profile.gauge
export = profile.export



from universe import error
import six

def merge_infos(info1, info2):
    """We often need to aggregate together multiple infos. Most keys can
    just be clobbered by the new info, but e.g. any keys which contain
    counts should be added. The merge schema is indicated by the key
    namespace.

    Namespaces:

    - stats.timers: Timing
    - stats.gauges: Gauge values
    - stats.*: Counts of a quantity
    """
    for key, value in six.iteritems(info2):
        if key in info1 and key.startswith('stats'):
            if key.startswith('stats.timers'):
                # timer
                info1[key] += value
            elif key.startswith('stats.gauges'):
                # gauge
                info1[key] = value
            else:
                # counter
                info1[key] += value
        else:
            info1[key] = value

def merge_reward_n(accum_reward_n, reward_n):
    for i in range(len(reward_n)):
        if reward_n[i] is not None:
            # Add rewards
            accum_reward_n[i] += reward_n[i]

def merge_done_n(accum_done_n, done_n):
    for i in range(len(done_n)):
        # Copy over done if the episode is indeed none
        if done_n[i]:
            accum_done_n[i] = done_n[i]

def _merge_observation(accum_observation, observation):
    """
    Old visual observation is discarded, because it is outdated frame.
    Text observations are merged, because they are messages sent from the rewarder.
    """
    if observation is None:
        # We're currently masking. So accum_observation probably
        # belongs to the previous episode. We may lose a "text"
        # observation from the previous episode, but that's ok.
        return None
    elif accum_observation is None:
        # Nothing to merge together
        return observation

    accum_observation['vision'] = observation.get('vision')
    accum_observation['text'] = accum_observation.get('text', []) + observation.get('text', [])
    return accum_observation

def merge_observation_n(accum_observation_n, observation_n):
    # Merge observations.
    for i in range(len(accum_observation_n)):
        accum_observation_n[i] = _merge_observation(accum_observation_n[i], observation_n[i])

def merge_n(
        accum_observation_n, accum_reward_n, accum_done_n, accum_info,
        observation_n, reward_n, done_n, info,
):
    # Merge observation/reward/done
    merge_observation_n(accum_observation_n, observation_n)
    merge_reward_n(accum_reward_n, reward_n)
    merge_done_n(accum_done_n, done_n)

    # Merge together infos. We deep merge the 'n' key and do a
    # simple merge on everything else.
    accum_info_n = accum_info['n']
    for accum_info_i, info_i in zip(accum_info_n, info['n']):
        merge_infos(accum_info_i, info_i)

    merge_infos(accum_info, info)
    accum_info['n'] = accum_info_n



# loaded inside of the environments

import logging
import os
from universe import pyprofile
import sys
is_py2 = sys.version[0] == '2'
if is_py2:
    import Queue as queue
else:
    import queue as queue
import threading
import time
import ujson
import collections

from autobahn.twisted import websocket
from universe.twisty import reactor

from universe import error, utils

logger = logging.getLogger(__name__)

class Exit(Exception):
    pass

class RewarderProtocol(websocket.WebSocketServerProtocol):
    connections = None

    def onConnect(self, request):
        if not os.path.exists('/usr/local/openai/privileged_state/password'):
            raise error.Error('No such file: /usr/local/openai/privileged_state/password. (HINT: did the init script run /app/universe-envs/base/openai-setpassword?)')
        with open('/usr/local/openai/privileged_state/password') as f:
            password = f.read().strip()

        self._message_id = 0
        self._request = request
        self._observer = request.headers.get('openai-observer') == 'true'
        self.password = password

        logger.info('Client connecting: peer=%s observer=%s', request.peer, self._observer)

    def authenticate(self, request):
        # Ugly, but it'll have to do for now.
        authorization = request.headers.get('authorization')
        if authorization is None:
            logger.info('REJECT REASON: No authorization header supplied: %s', request.headers)
            self.reject('No authorization header supplied. You must supply a basic authentication header.')
            return
        basic = utils.basic_auth_decode(authorization)
        if basic is None:
            logger.info('REJECT REASON: Invalid basic auth header: %s', request.headers)
            self.reject('Could not parse authorization header. You must supply a basic authentication header.')
            return
        username, password = basic
        if username != self.password:
            logger.info('REJECT REASON: Invalid password: %r (%s expected; %s)', username, self.password, request.headers)
            self.reject('Invalid password: {!r}. If you are using the allocator, you should see your password in the logs; if spinning up an environment by hand, it defaults to "openai". Connect as vnc://<ip>:<port>?password=<password>.'.format(username))
            return

    def onOpen(self):
        logger.info('WebSocket connection established')
        # Need to wait until onOpen to send messages
        self.authenticate(self._request)
        self.factory.agent_conn._register(self, observer=self._observer)

        # Inform the agent about the current env status
        env_info = self.factory.agent_conn.env_status.env_info()
        # Immediately upon connection, let the agent know the current
        # status.
        self.factory.agent_conn.send_env_describe_from_env_info(
            env_info,
        )

    def onMessage(self, payload, isBinary):
        if not self.factory.agent_conn.check_message(self):
            return

        assert not isBinary, "Binary websocket not supported"
        payload = ujson.loads(payload)

        context = {
            'start': time.time(),
            'conn': self,
        }
        latency = context['start'] - payload['headers']['sent_at']
        pyprofile.incr('rewarder_protocol.messages')
        pyprofile.incr('rewarder_protocol.messages.{}'.format(payload['method']))

        pyprofile.timing('rewarder_protocol.latency.rtt.skew_unadjusted', 2*latency)
        if latency < 0:
            pyprofile.incr('rewarder_protocol.latency.rtt.skew_unadjusted.negative')

        if payload['method'] == 'v0.env.reset':
            logger.info('Received reset message: %s', payload)
            self.factory.agent_conn.control_buffer.recv_rpc(context, payload)
        elif payload['method'] == 'v0.control.ping':
            logger.debug('Received ping message: %s', payload)
            parent_message_id = payload['headers']['message_id']
            headers = {'parent_message_id': parent_message_id}
            self.send_message('v0.reply.control.ping', {}, headers)
        else:
            logger.warn('Received unsupported message: %s', payload)

    def onClose(self, wasClean, code, reason):
        logger.info('WebSocket connection closed: %s', reason)
        self.factory.agent_conn._unregister(self)

    def send_message(self, method, body, headers):
        id = self._message_id

        self._message_id += 1
        new_headers = {
            'message_id': id,
            'sent_at': time.time(),
        }
        if headers:
            new_headers.update(headers)

        payload = {
            'method': method,
            'body': body,
            'headers': new_headers,
        }

        # This is a bit ugly, but decide how much we care
        if (method != 'v0.reply.control.ping' and 'parent_message_id' in new_headers) or\
           method == 'v0.connection.close':
            logger.info('Sending rewarder message: %s', payload)
        else:
            logger.debug('Sending rewarder message: %s', payload)

        self.sendMessage(ujson.dumps(payload).encode('utf-8'), False)

    def reject(self, message):
        self.send_message('v0.connection.close', {'message': message}, {})
        self.sendClose(code=1000, reason=message)
        self.transport.loseConnection()

class ControlBuffer(object):
    def __init__(self, cv):
        self.buf = queue.Queue()
        self.cv = cv

    def recv_rpc(self, context, payload):
        """Call from any thread"""
        logger.debug("Adding RPC payload to ControlBuffer queue: %s", payload)
        self.buf.put(('rpc', (context, payload)))
        with self.cv:
            self.cv.notifyAll()

    def client_disconnect(self, conn, stats):
        self.buf.put(('client_disconnect', (conn, stats)))

    def get(self, *args, **kwargs):
        """Call from main thread."""
        payload = self.buf.get(*args, **kwargs)
        logger.debug("Removing RPC payload from ControlBuffer queue: %s", payload)
        return payload

class AgentConn(object):
    def __init__(self, env_status, cv, control_buffer, error_buffer, idle_timeout=None, exclusive=False):
        self.error_buffer = error_buffer

        self.env_status = env_status
        self.control_buffer = control_buffer
        self.cv = cv
        self.conns = {}
        self.exclusive = exclusive

        self.idle_timeout = idle_timeout
        self.last_disconnect_time = time.time()
        self._idle_message_interval = 10 # for logging

    def active_clients(self):
        return [conn for conn, stats in self.conns.items() if stats['active']]

    def listen(self, port=15900):
        logger.info('Starting Rewarder on port=%s', port)
        factory = websocket.WebSocketServerFactory()
        factory.agent_conn = self
        factory.protocol = RewarderProtocol

        reactor.callFromThread(reactor.listenTCP, port, factory)

    def check_message(self, conn):
        with self.cv:
            self.conns[conn]['messages'] += 1
            if self.conns[conn]['active']:
                return True
            elif self.conns[conn]['observer']:
                logger.info('CONNECTION STATUS: Marking connection as active: observer=%s peer=%s total_conns=%d', True, conn._request.peer, len(self.conns))
                self.conns[conn]['active'] = True
                return True
            else:
                # Note: if this exceptions, Autobahn will end up capturing
                # the errors since its hooks are called via
                # maybeDeferred. This won't always print the prettiest
                # stack trace.

                # This conn is neither active or an observer - before setting
                # active, let's see if there are any existing active,
                # non-observer conns.
                active = len([o for o in self.conns.values() if not o['observer'] and o['active']])
                if active > 0:
                    # Already full up, sorry!
                    logger.info('CONNECTION STATUS: Dropping new connection since already have %d non-observer conns (%d conns total)', active, len(self.conns))
                    # Sometimes connecting clients will time out before
                    # the connection is fully established, but the
                    # rewarder would still count the session as having
                    # come up. This, we try to wait until as late as
                    # possible to decide if a client is active.
                    conn.reject('The rewarder already has an active client. (HINT: if you obtained your environment through the allocator, make sure to run .configure(client_id=...) with a different client_id for each concurrent worker.)')
                    return False
                else:
                    logger.info('CONNECTION STATUS: Marking connection as active: observer=%s peer=%s total_conns=%d', False, conn._request.peer, len(self.conns))
                    self.conns[conn]['active'] = True
                    return True

    def _register(self, conn, observer=False):
        with self.cv:
            self.conns[conn] = {'messages': 0, 'observer': observer, 'active': False}
            self.cv.notifyAll()

    def _unregister(self, conn):
        with self.cv:
            try:
                stats = self.conns.pop(conn)
            except KeyError:
                stats = None
            else:
                self.cv.notifyAll()

            if stats is not None and stats['active']:
                self.last_disconnect_time = time.time()
                active = self.active_clients()
                logger.info('[%s] Active client disconnected (sent %d messages). Still have %d active clients left', utils.thread_name(), stats['messages'], len(active))
            else:
                logger.info('[%s] Non-active client disconnected', utils.thread_name())

            self.control_buffer.client_disconnect(conn, stats)

    def _broadcast(self, method, body, headers=None, conn=None):
        if conn:
            conns = [conn]
        else:
            conns = self.conns

        for conn in conns:
            conn.send_message(method, body, headers)

    def send_env_text(self, text, episode_id):
        ''' text channel to communicate with the agent '''
        reactor.callFromThread(self._send_env_text, text, episode_id)

    def _send_env_text(self, text, episode_id):
        self._broadcast('v0.env.text', {
            'text': text
        }, {'episode_id': episode_id})

    def send_env_observation(self, observation, episode_id):
        reactor.callFromThread(self._send_env_observation, observation, episode_id)

    def _send_env_observation(self, observation, episode_id, conn=None):
        self._broadcast('v0.env.observation', {
            'observation': observation,
        }, {'episode_id': episode_id}, conn=conn)

    def send_env_reward(self, reward, done, info, episode_id):
        pyprofile.incr('agent_conn.reward', reward)
        if done:
            pyprofile.incr('agent_conn.done')

        reactor.callFromThread(self._send_env_reward, reward, done, info, episode_id)

    def _send_env_reward(self, reward, done, info, episode_id):
        self._broadcast('v0.env.reward', {
            'reward': reward,
            'done': done,
            'info': info,
        }, {'episode_id': episode_id})

    def send_env_describe_from_env_info(self, env_info):
        assert env_info['fps'] is not None, "Missing fps: {}".format(env_info)
        self.send_env_describe(env_info['env_id'], env_info['env_state'], episode_id=env_info['episode_id'], fps=env_info['fps'])

    def send_env_describe(self, env_id, env_state, episode_id, fps, headers=None, parent_message_id=None, parent_context=None):
        reactor.callFromThread(self._send_env_describe, env_id, env_state, episode_id, fps, headers, parent_message_id, parent_context)

    def _send_env_describe(self, env_id, env_state, episode_id, fps, headers=None, parent_message_id=None, parent_context=None):
        conn = None
        if headers is None:
            headers = {}

        if parent_message_id is not None:
            headers['parent_message_id'] = parent_message_id
            headers['parent_runtime'] = time.time() - parent_context['start']
            conn = parent_context['conn']

        headers['episode_id'] = episode_id

        assert fps is not None
        # TODO: decide how to handle multiple concurrent envs
        self._broadcast('v0.env.describe', {
            'env_id': env_id,
            'env_state': env_state,
            'fps': fps,
        }, headers, conn)

    def send_reply_error(self, *args, **kwargs):
        reactor.callFromThread(self._send_reply_error, *args, **kwargs)

    def _send_reply_error(self, message, parent_message_id, parent_context):
        headers = {}
        headers['parent_message_id'] = parent_message_id
        headers['parent_runtime'] = time.time() - parent_context['start']
        conn = parent_context['conn']

        # TODO: decide how to handle multiple concurrent envs
        self._broadcast('v0.reply.error', {
            'message': message,
        }, headers, conn)

    def send_reply_env_reset(self, *args, **kwargs):
        reactor.callFromThread(self._send_reply_env_reset, *args, **kwargs)

    def _send_reply_env_reset(self, parent_message_id, parent_context, episode_id):
        headers = {}
        headers['parent_message_id'] = parent_message_id
        headers['parent_runtime'] = time.time() - parent_context['start']
        headers['episode_id'] = episode_id
        conn = parent_context['conn']

        # TODO: decide how to handle multiple concurrent envs
        self._broadcast('v0.reply.env.reset', {}, headers, conn)

    def check_status(self):
        with self.cv:
            if self.idle_timeout is None:
                return

            active = self.active_clients()
            if len(active) == 0:
                now = time.time()
                idle_duration = now - self.last_disconnect_time

                if self.idle_timeout is not None:
                    utils.periodic_log(
                        self, 'idle_timeout',
                        'No active clients for %.2fs (total client: %d); will exit due to idle timeout after %.0fs',
                        idle_duration, len(self.conns), self.idle_timeout, frequency=self._idle_message_interval)
                else:
                    utils.periodic_log(
                        self, 'idle_for',
                        'No active clients for %.2fs (total client: %d)',
                        idle_duration, len(self.conns), frequency=self._idle_message_interval)

                if self.idle_timeout is not None and idle_duration > self.idle_timeout:
                    self.error_buffer.record(Exit('EXIT CAUSE: idle timeout exceeded after {:.2f} seconds'.format(idle_duration)), wrap=False)

class RewardLogger(object):
    def __init__(self):
        self.reset(log=False, episode_stats=True)

    def reset(self, log=True, episode_stats=True):
        if log:
            self._log()

        self.last_print = time.time()
        # Could just maintain summary statistics, but we're not going
        # to have more than fps rewards at once, so it's fine to just
        # keep them all around.
        self.reward = []
        self.done = False
        self.info = {}
        self.count = 0

        if episode_stats:
            if log:
                self._log_reset()
            self.episode_reward = 0
            self.episode_count = 0
            self.episode_start = time.time()

    def record(self, reward, done, info):
        self.reward.append(reward)
        self.done = self.done or done
        self.info.update(info)
        self.count += 1

        self.episode_reward += reward
        self.episode_count += 1

        if time.time() - self.last_print > 1:
            self._log()
            self.reset(log=False, episode_stats=False)

    def _log_reset(self):
        logger.info('[%s] Ending previous episode: episode_reward=%s episode_count=%s episode_duration=%.2f', utils.thread_name(), self.episode_reward, self.episode_count, time.time() - self.episode_start)

    def _log(self):
        if 'rewarder.profile' in self.info:
            self.info['rewarder.profile'] = '<{} bytes>'.format(len(str(self.info['rewarder.profile'])))

        if len(self.reward) > 0:
            min_reward = min(self.reward)
            max_reward = max(self.reward)
        else:
            min_reward = '(empty)'
            max_reward = '(empty)'

        logger.info('[%s] Over past %.2fs, sent %s reward messages to agent: reward=%s reward_min=%s reward_max=%s done=%s info=%s',
                    utils.thread_name(), time.time() - self.last_print, self.count,
                    sum(self.reward), min_reward, max_reward,
                    self.done, self.info)



import six
import uuid

from universe import utils
from universe.vncdriver.vendor import pydes

class RFBDes(pydes.des):
    def setKey(self, key):
        key = key.encode('ascii')

        newkey = []
        for ki in range(len(key)):
            if six.PY2:
                bsrc = ord(key[ki])
            else:
                bsrc = key[ki]

            # Reverse the bits
            btgt = 0
            for i in range(8):
                if bsrc & (1 << i):
                    btgt = btgt | (1 << 7-i)

            if six.PY2:
                newkey.append(chr(btgt))
            else:
                newkey.append(btgt)

        super(RFBDes, self).setKey(newkey)

def challenge():
    length = 16
    buf = b''
    while len(buf) < length:
        entropy = uuid.uuid4().bytes
        buf += entropy
    return buf[:length]

def challenge_response(challenge, password=None):
    if password is None:
        password = utils.default_password()
    password += ((8 - len(password)) % 8) * '\0'  # pad to multiple of 8 bytes
    des = RFBDes(password)
    return des.encrypt(challenge)



import string
from universe import error
from universe.vncdriver import constants

class VNCEvent(object):
    pass

def keycode(key):
    if key in constants.KEYMAP:
        return constants.KEYMAP.get(key)
    elif len(key) == 1:
        return ord(key)
    else:
        raise error.Error('Not sure how to translate to keycode: {!r}'.format(key))

class KeyEvent(VNCEvent):
    _keysym_to_name = {}
    for key, value in constants.KEYMAP.items():
        _keysym_to_name[value] = key
    for c in string.printable:
        _keysym_to_name[ord(c)] = c

    @classmethod
    def build(cls, keys, down=None):
        """Build a key combination, such as:

        ctrl-t
        """
        codes = []
        for key in keys.split('-'):
            key = keycode(key)
            codes.append(key)

        events = []
        if down is None or down:
            for code in codes:
                events.append(cls(code, down=True))

        if down is None or not down:
            for code in reversed(codes):
                events.append(cls(code, down=False))
        return events

    @classmethod
    def by_name(cls, key, down=None):
        return cls(keycode(key), down=down)

    def __init__(self, key, down=True):
        # TODO: validate key
        self.key = key
        self.down = bool(down)

    def compile(self):
        return 'KeyEvent', self.key, self.down

    def __repr__(self):
        if self.down:
            direction = 'down'
        else:
            direction = 'up'
        name = self._keysym_to_name.get(self.key)
        if not name:
            name = '0x{:x}'.format(self.key)
        else:
            name = '{} (0x{:x})'.format(name, self.key)
        return 'KeyEvent<key={} direction={}>'.format(name, direction)

    def __str__(self):
        return repr(self)

    def __hash__(self):
        return (self.key, self.down).__hash__()

    def __eq__(self, other):
        return type(other) == type(self) and \
            other.key == self.key and \
            other.down == self.down

    @property
    def key_name(self):
        """Human readable name"""
        return self._keysym_to_name.get(self.key)

class PointerEvent(VNCEvent):
    def __init__(self, x, y, buttonmask=0):
        self.x = x
        self.y = y
        self.buttonmask = buttonmask

    def compile(self):
        return 'PointerEvent', self.x, self.y, self.buttonmask

    def __repr__(self):
        return 'PointerEvent<x={} y={} buttonmask={}>'.format(self.x, self.y, self.buttonmask)

    def __str__(self):
        return repr(self)



import logging
import multiprocessing
import numpy as np
import traceback

import gym
from gym import spaces
from universe.vectorized import core

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class Error(Exception):
    pass

def display_name(exception):
    prefix = ''
    # AttributeError has no __module__; RuntimeError has module of
    # exceptions
    if hasattr(exception, '__module__') and exception.__module__ != 'exceptions':
        prefix = exception.__module__ + '.'
    return prefix + type(exception).__name__

def render_dict(error):
    return {
        'type': display_name(error),
        'message': error.message,
        'traceback': traceback.format_exc(error)
    }

class Worker(object):
    def __init__(self, env_m, worker_idx):
        # These are instantiated in the *parent* process
        # currently. Probably will want to change this. The parent
        # does need to obtain the relevant Spaces at some stage, but
        # that's doable.
        self.worker_idx = worker_idx
        self.env_m = env_m
        self.m = len(env_m)
        self.parent_conn, self.child_conn = multiprocessing.Pipe()
        self.joiner = multiprocessing.Process(target=self.run)
        self._clear_state()

        self.start()

        # Parent only!
        self.child_conn.close()

    def _clear_state(self):
        self.mask = [True] * self.m

    # Control methods

    def start(self):
        self.joiner.start()

    def _parent_recv(self):
        rendered, res = self.parent_conn.recv()
        if rendered is not None:
            raise Error('[Worker {}] Error: {} ({})\n\n{}'.format(self.worker_idx, rendered['message'], rendered['type'], rendered['traceback']))
        return res

    def _child_send(self, msg):
        self.child_conn.send((None, msg))

    def _parent_send(self, msg):
        try:
            self.parent_conn.send(msg)
        except IOError: # the worker is now dead
            try:
                res = self._parent_recv()
            except EOFError:
                raise Error('[Worker {}] Child died unexpectedly'.format(self.worker_idx))
            else:
                raise Error('[Worker {}] Child returned unexpected result: {}'.format(self.worker_idx, res))

    def close_start(self):
        self._parent_send(('close', None))

    def close_finish(self):
        self.joiner.join()

    def reset_start(self):
        self._parent_send(('reset', None))

    def reset_finish(self):
        return self._parent_recv()

    def step_start(self, action_m):
        """action_m: the batch of actions for this worker"""
        self._parent_send(('step', action_m))

    def step_finish(self):
        return self._parent_recv()

    def mask_start(self, i):
        self._parent_send(('mask', i))

    def seed_start(self, seed_m):
        self._parent_send(('seed', seed_m))

    def render_start(self, mode, close):
        self._parent_send(('render', (mode, close)))

    def render_finish(self):
        return self._parent_recv()

    def run(self):
        try:
            self.do_run()
        except Exception as e:
            rendered = render_dict(e)
            self.child_conn.send((rendered, None))
            return

    def do_run(self):
        # Child only!
        self.parent_conn.close()

        while True:
            method, body = self.child_conn.recv()
            logger.debug('[%d] Received: method=%s body=%s', self.worker_idx, method, body)
            if method == 'close':
                logger.info('Closing envs')
                # TODO: close envs?
                return
            elif method == 'reset':
                self._clear_state()
                observation_m = [env.reset() for env in self.env_m]
                self._child_send(observation_m)
            elif method == 'step':
                action_m = body
                observation_m, reward_m, done_m, info = self.step_m(action_m)
                self._child_send((observation_m, reward_m, done_m, info))
            elif method == 'mask':
                i = body
                assert 0 <= i < self.m, 'Bad value for mask: {} (should be >= 0 and < {})'.format(i, self.m)

                self.mask[i] = False
                logger.debug('[%d] Applying mask: i=%d', self.worker_idx, i)
            elif method == 'seed':
                seeds = body
                [env.seed(seed) for env, seed in zip(self.env_m, seeds)]
            elif method == 'render':
                mode, close = body
                if mode == 'human':
                    self.env_m[0].render(mode=mode, close=close)
                    result = [None]
                else:
                    result = [env.render(mode=mode, close=close) for env in self.env_m]
                self._child_send(result)
            else:
                raise Error('Bad method: {}'.format(method))

    def step_m(self, action_m):
        observation_m = []
        reward_m = []
        done_m = []
        info = {'m': []}

        for env, enabled, action in zip(self.env_m, self.mask, action_m):
            if enabled:
                observation, reward, done, info_i = env.step(action)
                if done:
                    observation = env.reset()
            else:
                observation = None
                reward = 0
                done = False
                info_i = {}
            observation_m.append(observation)
            reward_m.append(reward)
            done_m.append(done)
            info['m'].append(info_i)
        return observation_m, reward_m, done_m, info


def step_n(worker_n, action_n):
    accumulated = 0
    for worker in worker_n:
        action_m = action_n[accumulated:accumulated+worker.m]
        worker.step_start(action_m)
        accumulated += worker.m

    observation_n = []
    reward_n = []
    done_n = []
    info = {'n': []}

    for worker in worker_n:
        observation_m, reward_m, done_m, info_i = worker.step_finish()
        observation_n += observation_m
        reward_n += reward_m
        done_n += done_m
        info['n'] += info_i['m']
    return observation_n, reward_n, done_n, info


def reset_n(worker_n):
    for worker in worker_n:
        worker.reset_start()

    observation_n = []
    for worker in worker_n:
        observation_n += worker.reset_finish()

    return observation_n


def seed_n(worker_n, seed_n):
    accumulated = 0
    for worker in worker_n:
        action_m = seed_n[accumulated:accumulated+worker.m]
        worker.seed_start(seed_n)
        accumulated += worker.m


def mask(worker_n, i):
    accumulated = 0
    for k, worker in enumerate(worker_n):
        if accumulated + worker.m <= i:
            accumulated += worker.m
        else:
            worker.mask_start(i - accumulated)
            return

def render_n(worker_n, mode, close):
    if mode == 'human':
        # Only render 1 worker
        worker_n = worker_n[0:]

    for worker in worker_n:
        worker.render_start(mode, close)
    res = []
    for worker in worker_n:
        res += worker.render_finish()
    if mode != 'human':
        return res
    else:
        return None

def close_n(worker_n):
    if worker_n is None:
        return

    # TODO: better error handling: workers should die when we go away
    # anyway. Also technically should wait for these processes if
    # we're not crashing.
    for worker in worker_n:
        try:
            worker.close_start()
        except Error:
            pass

    # for worker in worker_n:
    #     try:
    #         worker.close_finish()
    #     except Error:
    #         pass

class MultiprocessingEnv(core.Env):
    metadata = {
        'runtime.vectorized': True,
    }

    def __init__(self, env_id):
        self.worker_n = None

        # Pull the relevant info from a transient env instance
        self.spec = gym.spec(env_id)
        env = self.spec.make()

        current_metadata = self.metadata
        self.metadata = env.metadata.copy()
        self.metadata.update(current_metadata)

        self.action_space = env.action_space
        self.observation_space = env.observation_space
        self.reward_range = env.reward_range

    def _configure(self, n=1, pool_size=None, episode_limit=None):
        super(MultiprocessingEnv, self)._configure()
        self.n = n
        self.envs = [self.spec.make() for _ in range(self.n)]

        if pool_size is None:
            pool_size = min(len(self.envs), multiprocessing.cpu_count() - 1)
            pool_size = max(1, pool_size)

        self.worker_n = []
        m = int((self.n + pool_size - 1) / pool_size)
        for i in range(0, self.n, m):
            envs = self.envs[i:i+m]
            self.worker_n.append(Worker(envs, i))

        if episode_limit is not None:
            self._episode_id.episode_limit = episode_limit

    def _seed(self, seed):
        seed_n(self.worker_n, seed)
        return [[seed_i] for seed_i in seed]

    def _reset(self):
        return reset_n(self.worker_n)

    def _step(self, action_n):
        return step_n(self.worker_n, action_n)

    def _render(self, mode='human', close=False):
        return render_n(self.worker_n, mode=mode, close=close)

    def mask(self, i):
        mask(self.worker_n, i)

    def _close(self):
        close_n(self.worker_n)

if __name__ == '__main__':
    env_n = make('Pong-v3')
    env_n.configure()
    env_n.reset()
    print(env_n.step([0] * 10))



import logging
import threading

logger = logging.getLogger()

def parse_episode_id(episode_id):
    if episode_id is None:
        return -1
    return int(episode_id)

def generate_episode_id(parsed):
    if parsed == -1:
        return None
    return str(parsed)

def compare_ids(a, b):
    if a == b:
        return 0
    elif a is None:
        return -1
    elif b is None:
        return 1
    elif parse_episode_id(a) < parse_episode_id(b):
        return -1
    else:
        return 1

class EnvStatus(object):
    def __init__(self, label=None, primary=True):
        self.cv = threading.Condition()
        self._env_id = None
        self._env_state = None
        self._episode_id = '0'
        self._fps = None
        self.label = label or 'EnvStatus'
        self.primary = primary

    def env_info(self):
        with self.cv:
            return {
                'env_state': self._env_state,
                'env_id': self._env_id,
                'episode_id': self._episode_id,
                'fps': self._fps,
            }

    def set_env_info(self, env_state=None, env_id=None, episode_id=None, bump_past=None, fps=None):
        """Atomically set the environment state tracking variables.
        """
        with self.cv:
            if env_id is None:
                env_id = self._env_id
            if env_state is None:
                env_state = self._env_state
            if fps is None:
                fps = self._fps
            self.cv.notifyAll()

            old_episode_id = self._episode_id
            if self.primary:
                current_id = parse_episode_id(self._episode_id)
                # Bump when changing from resetting -> running
                if bump_past is not None:
                    bump_past_id = parse_episode_id(bump_past)
                    current_id = max(bump_past_id+1, current_id+1)
                elif env_state == 'resetting':
                    current_id += 1
                self._episode_id = generate_episode_id(current_id)
                assert self._fps or fps
            elif episode_id is False:
                # keep the same episode_id: this is just us proactive
                # setting the state to resetting after a done=True
                pass
            else:
                assert episode_id is not None, "No episode_id provided. This likely indicates a misbehaving server, which did not send an episode_id"
                self._episode_id = episode_id
            self._fps = fps
            logger.info('[%s] Changing env_state: %s (env_id=%s) -> %s (env_id=%s) (episode_id: %s->%s, fps=%s)', self.label, self._env_state, self._env_id, env_state, env_id, old_episode_id, self._episode_id, self._fps)
            self._env_state = env_state
            if env_id is not None:
                self._env_id = env_id

            return self.env_info()

    @property
    def episode_id(self):
        with self.cv:
            return self._episode_id

    @property
    def env_state(self):
        with self.cv:
            return self._env_state

    @env_state.setter
    def env_state(self, value):
        # TODO: Validate env_state
        self.set_env_info(value)

    @property
    def env_id(self):
        with self.cv:
            return self._env_id

    @env_id.setter
    def env_id(self, value):
        self.set_env_info(None, env_id=value)

    @property
    def fps(self):
        with self.cv:
            return self._fps

    def wait_for_env_state_change(self, start_state):
        with self.cv:
            while True:
                if self._env_state != start_state:
                    return self.env_info()
                self.cv.wait(timeout=10)



import logging
import six
import sys
if six.PY2:
    import Queue as queue
else:
    import queue
import threading
from twisted.internet import defer

from universe.twisty import reactor

logger = logging.getLogger(__name__)

class ErrorBuffer(object):
    def __init__(self):
        self.queue = queue.Queue()

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        if value is not None:
            self.record(value)

    def __call__(self, error, wrap=True):
        self.record(error, wrap=True)

    def record(self, error, wrap=True):
        logger.debug('Error in thread %s: %s', threading.current_thread().name, error)
        if wrap:
            error = format_error(error)

        try:
            self.queue.put_nowait(error)
        except queue.Full:
            pass

    def check(self, timeout=None):
        if timeout is None:
            timeout = 0

        try:
            error = self.queue.get(timeout=timeout)
        except queue.Empty:
            return
        else:
            raise error

    def blocking_check(self, timeout=None):
        # TODO: get rid of this method
        if timeout is None:
            while True:
                self.check(timeout=3600)
        else:
            self.check(timeout)


from twisted.python import failure
import traceback
import threading
from universe import error
def format_error(e):
    # errback automatically wraps everything in a Twisted Failure
    if isinstance(e, failure.Failure):
        e = e.value

    if isinstance(e, str):
        err_string = e
    elif six.PY2:
        err_string = traceback.format_exc(e).rstrip()
    else:
        err_string = ''.join(traceback.format_exception(type(e), e, e.__traceback__)).rstrip()

    if err_string == 'None':
        # Reasonable heuristic for exceptions that were created by hand
        last = traceback.format_stack()[-2]
        err_string = '{}\n  {}'.format(e, last)
    # Quick and dirty hack for now.
    err_string = err_string.replace('Connection to the other side was lost in a non-clean fashion', 'Connection to the other side was lost in a non-clean fashion (HINT: this generally actually means we got a connection refused error. Check that the remote is actually running.)')
    return error.Error(err_string)

def queue_get(local_queue):
    while True:
        try:
            result = local_queue.get(timeout=1000)
        except queue.Empty:
            pass
        else:
            return result

def blockingCallFromThread(f, *a, **kw):
    local_queue = queue.Queue()
    def _callFromThread():
        result = defer.maybeDeferred(f, *a, **kw)
        result.addBoth(local_queue.put)
    reactor.callFromThread(_callFromThread)
    result = queue_get(local_queue)
    if isinstance(result, failure.Failure):
        if result.frames:
            e = error.Error(str(result))
        else:
            e = result.value
        raise e
    return result

from gym import spaces
def repeat_space(space, n):
    return spaces.Tuple([space] * n)

import base64
import uuid
def random_alphanumeric(length=14):
    buf = []
    while len(buf) < length:
        entropy = base64.encodestring(uuid.uuid4().bytes).decode('ascii')
        bytes = [c for c in entropy if c.isalnum()]
        buf += bytes
    return ''.join(buf)[:length]


def best_effort(function, *args, **kwargs):
    try:
        return function(*args, **kwargs)
    except:
        if six.PY2:
            logging.error('Error in %s:', function.__name__)
            traceback.print_exc()
        else:
            logging.error('Error in %s:', function.__name__)
            logger.error(traceback.format_exc())
        return None

import base64
def basic_auth_encode(username, password=''):
    fmt = '{}:{}'.format(username, password)
    return 'Basic ' + base64.encodestring(fmt.encode('utf-8')).rstrip().decode('utf-8')

def basic_auth_decode(header):
    if header.startswith('Basic '):
        header = header[len('Basic '):]
        decoded = base64.decodestring(header.encode('utf-8')).decode('utf-8')
        username, password = decoded.split(':')
        return username, password
    else:
        return None

import os
def default_password():
    if os.path.exists('/usr/local/openai/privileged_state/password'):
        with open('/usr/local/openai/privileged_state/password') as f:
            return f.read().strip()
    return 'openai'

import logging
import time
logger = logging.getLogger(__name__)
class PeriodicLog(object):
    def log(self, obj, name, msg, *args, **kwargs):
        try:
            info = obj._periodic_log_info
        except AttributeError:
            info = obj._periodic_log_info = {}

        # Would be better to use a frequency=... arg after kwargs, but
        # that isn't py2 compatible.
        frequency = kwargs.pop('frequency', 1)
        delay = kwargs.pop('delay', 0)
        last_log = info.setdefault(name, time.time()-frequency+delay)
        if time.time() - last_log < frequency:
            return
        info[name] = time.time()
        logger.info('[{}] {}'.format(name, msg), *args)

    def log_debug(self, obj, name, msg, *args, **kwargs):
        try:
            info = obj._periodic_log_debug
        except AttributeError:
            info = obj._periodic_log_debug = {}

        frequency = kwargs.pop('frequency', 1)
        delay = kwargs.pop('delay', 0)
        last_log = info.setdefault(name, time.time()-frequency+delay)
        if time.time() - last_log < frequency:
            return
        info[name] = time.time()
        logger.debug('[{}] {}'.format(name, msg), *args)

_periodic = PeriodicLog()
periodic_log = _periodic.log
periodic_log_debug = _periodic.log_debug

import threading
def thread_name():
    return threading.current_thread().name



import logging

from twisted.internet import defer, endpoints

from universe import error, utils
from universe.twisty import reactor
from universe.vncdriver import screen, vnc_client

logger = logging.getLogger(__name__)

class VNCSession(object):
    def __init__(self, remotes, error_buffer):
        self.remotes = remotes
        self.error_buffer = error_buffer
        self._pyglet_screen = None
        self.connect()

    def connect(self):
        utils.blockingCallFromThread(self._connect)

    def _connect(self):
        deferreds = []

        for i, remote in enumerate(self.remotes):
            d = defer.Deferred()
            deferreds.append(d)

            factory = vnc_client.client_factory(d, self.error_buffer)
            factory.rewarder_session = self
            factory.label = 'vnc:{}:{}'.format(i, remote)
            endpoint = endpoints.clientFromString(reactor, 'tcp:'+remote)

            def success(i):
                logger.info('[%s] VNC connection established', factory.label)

            def fail(reason):
                reason = error.Error('[{}] Connection failed: {}'.format(factory.label, reason.value))
                try:
                    d.errback(utils.format_error(reason))
                except defer.AlreadyCalledError:
                    pass
            endpoint.connect(factory).addCallback(success).addErrback(fail)

        d = defer.DeferredList(deferreds, fireOnOneErrback=True)

        def success(results):
            # Store the _clients list when connected
            self._clients = [client for success, client in results]
        d.addCallback(success)
        return d

    def flip(self):
        observation_n = []
        info_n = []
        for i, client in enumerate(self._clients):
            observation, info = client.numpy_screen.flip()
            updates = info['vnc_session.framebuffer_updates']

            # Keep the pyglet screen fed, but don't flip it until the user calls render
            if i == 0 and self._pyglet_screen:
                for update in updates:
                    self._pyglet_screen.apply(update)

            observation_n.append(observation)
            info_n.append({'vnc.updates.n': len(updates)})

        return observation_n, info_n

    def peek(self):
        observations = [client.numpy_screen.peek() for client in self._clients]
        return observations

    def step(self, action):
        reactor.callFromThread(self._step, action)
        return self.flip()

    def _step(self, action):
        try:
            for a, client in zip(action, self._clients):
                for event in a:
                    if event[0] == 'KeyEvent':
                        key, down = event[1:]
                        client.send_KeyEvent(key, down)
                    elif event[0] == 'PointerEvent':
                        x, y, buttomask = event[1:]
                        client.send_PointerEvent(x, y, buttomask)
                    else:
                        raise error.Error('Bad event type: {}'.format(type))
        except Exception as e:
            self.error_buffer.record(e)

    def render(self):
        if not self._pyglet_screen:
            start = self.peek()[0]
            self._pyglet_screen = screen.PygletScreen(start)
        self._pyglet_screen.flip()

    def close(self):
        utils.blockingCallFromThread(self._close)

    def _close(self):
        if getattr(self, '_clients', None) is not None:
            for client in self._clients:
                client.close()
            self._clients = None



import json
import logging
import os
import time

from autobahn.twisted import websocket
from universe.twisty import reactor
from twisted.internet import endpoints

logger = logging.getLogger(__name__)

class RewardServerClient(websocket.WebSocketClientProtocol, object):
    def __init__(self):
        super(RewardServerClient, self).__init__()
        self.id = -1

        self.proxy_server = None
        self._connected = False

    def onConnect(self, request):
        self.id = self.factory.proxy_server.id
        logger.info('[RewardProxyClient] [%d] Connected to rewarder', self.id)
        self.proxy_server = self.factory.proxy_server
        self._connected = True

        buffered = self.proxy_server.pop_buffer()
        logger.info('[RewardProxyClient] [%d] Flushing %d buffered messages', self.id, len(buffered))
        for msg in buffered:
            self.sendMessage(msg)

    def onOpen(self):
        logger.info('[RewardProxyClient] [%d] Rewarder websocket connection established', self.id)

    def onMessage(self, msg, isBinary):
        logger.debug('[RewardProxyClient] [%d] Received message from server: %s', self.id, msg)
        self.proxy_server.sendMessage(msg)

        # Record the message
        self.proxy_server.record_message(msg.decode('utf-8'), from_rewarder=True)

        # # Process the message for recording
        # method, headers, body = unpack_message(msg)
        #
        # if method == "env.reward":
        #     # {"body":{"info":{"episode":0},"reward":0.0,"done":false},
        #     # "headers":{"sent_at":1473126129.231828928,"message_id":207},
        #     # "method":"env.reward"}

    def onClose(self, wasClean, code, reason):
        logger.info('[RewardProxyClient] [%d] Rewarder websocket connection closed: %s', self.id, reason)

    def close(self):
        logger.info('[RewardProxyClient] [%d] Closing connection', self.id)
        self.transport.loseConnection()


class RewardProxyServer(websocket.WebSocketServerProtocol, object):
    _next_id = 0
    _n_open_files = 0

    @classmethod
    def next_id(cls):
        id = cls._next_id
        cls._next_id += 1
        return id

    def __init__(self):
        super(RewardProxyServer, self).__init__()
        self.id = self.next_id()
        self.client = None
        self.file = None  # We do not open open the file until we have established an end-to-end connection
        self.buffered = []

        self._closed = False

    def pop_buffer(self):
        """Called by the client once it's ready to start sending messages.
        """
        buffered = self.buffered
        self.buffered = []
        return buffered

    def begin_recording(self):
        """
        Open the file and write the metadata header to describe this recording. Called after we establish an end-to-end connection
        This uses Version 1 of our protocol

        Version 0 can be seen here: https://github.com/openai/universe/blob/f85a7779c3847fa86ec7bb513a1da0d3158dda78/bin/recording_agent.py
        """
        logger.info("[RewardProxyServer] [%d] Starting recording", self.id)

        if self._closed:
            logger.error(
                "[RewardProxyServer] [%d] Attempted to start writing although client connection is already closed. Aborting", self.id)
            self.close()
            return

        if self._n_open_files != 0:
            logger.error("[RewardProxyServer] [%d] WARNING: n open rewards files = %s. This is unexpected. Dropping connection.", self.id, self._n_open_files)
            self.close()
            return

        logfile_path = os.path.join(self.factory.logfile_dir, 'rewards.demo')
        logger.info('Recording to {}'.format(logfile_path))
        self.file = open(logfile_path, 'w')

        self._n_open_files += 1
        logger.info("[RewardProxyServer] [%d] n open rewards files incremented: %s", self.id, self._n_open_files)

        self.file.write(json.dumps({
            'version': 1,
            '_debug_version': '0.0.1',  # Give this an internal version for debugging corrupt reward.demo files # TODO, pull this from setup.py or the host docker image
        }))
        self.file.write('\n')
        self.file.flush()

        logger.info("[RewardProxyServer] [%d] Wrote version number", self.id)

    def onConnect(self, request):
        logger.info('[RewardProxyServer] [%d] Client connecting: %s', self.id, request.peer)
        self._request = request

    def onOpen(self):
        logger.info("[RewardProxyServer] [%d] Websocket connection established", self.id)
        self.connect_upstream()

    def connect_upstream(self, tries=1, max_attempts=7):
        if self._closed:
            logger.info("[RewardProxyServer] [%d] Attempted to connect upstream although client connection is already closed. Aborting",
                        self.id)
            return

        remote = getattr(self.factory, 'rewarder_address', 'localhost:15900')
        endpoint = endpoints.clientFromString(reactor, 'tcp:' + remote)
        client_factory = websocket.WebSocketClientFactory('ws://' + remote)
        headers = {'authorization': self._request.headers['authorization']}
        if self._request.headers.get('openai-observer'):
            headers['openai-observer'] = self._request.headers.get('openai-observer')
        client_factory.headers = headers
        client_factory.protocol = RewardServerClient
        client_factory.proxy_server = self
        client_factory.endpoint = endpoint

        logger.info("[RewardProxyServer] [%d] Connecting to upstream %s (try %d/%d)", self.id, remote, tries, max_attempts)

        def _connect_callback(client):
            logger.info('[RewardProxyServer] [%d] Upstream connection %s established', self.id, remote)
            self.client = client
            if self.factory.logfile_dir:
                self.begin_recording()

        def _connect_errback(reason):
            if tries < max_attempts:
                # Somewhat arbitrary exponential backoff: should be
                # pretty rare, and indicate that we're just starting
                # up.
                delay = 1.5 ** tries
                logger.info('[RewardProxyServer] [%d] Connection to %s failed: %s. Try %d/%d; going to retry in %fs', self.id, remote, reason, tries, max_attempts, delay)
                reactor.callLater(
                    delay, self.connect_upstream,
                    tries=tries+1, max_attempts=max_attempts)
            else:
                logger.error('[RewardProxyServer] [%d] Connection to %s failed: %s. Completed %d/%d atttempts; disconnecting.', self.id, remote, reason, tries, max_attempts)
                self.transport.loseConnection()

        endpoint.connect(client_factory).addCallbacks(_connect_callback, _connect_errback)

    def close(self):
        logger.info('[RewardProxyServer] [%d] Closing...', self.id)
        self.transport.loseConnection()

    def onClose(self, wasClean, code, reason):
        logger.info('[RewardProxyServer] [%d] Client connection closed: %s', self.id, reason)
        if self.client:
            self.client.close()
        if self.file:
            self.file.close()

        self._closed = True

    def onMessage(self, msg, binary):
        logger.debug('[RewardProxyServer] [%d] Received message from client: %s', self.id, msg)

        # Pass the message on to the client
        if self.client and self.client._connected:
            self.client.sendMessage(msg)
        else:
            self.buffered.append(msg)

        self.record_message(msg.decode('utf-8'), from_rewarder=False)

    def record_message(self, msg, from_rewarder):
        """Record a message to our rewards.demo file if it is has been opened"""
        if self.file:
            # Include an authoritative timestamp (because the `sent_at` from the server is likely to be different
            timestamped_message = {
                'timestamp': time.time(),
                'message': json.loads(msg),
                'from_rewarder': from_rewarder,
            }
            self.file.write(json.dumps(timestamped_message))
            self.file.write('\n')
            self.file.flush()



import logging
import threading
import time

from universe import error
from universe.rewarder import env_status, merge

logger = logging.getLogger(__name__)
extra_logger = logging.getLogger('universe.extra.'+__name__)

class RewardState(object):
    def __init__(self, label, episode_id):
        self.label = label

        self.count = 0
        self.reward = 0.
        self.text = []
        self.done = False
        self.info = {}
        self._observation = None

        self._episode_id = episode_id
        self._env_state = None

    def push_time(self, remote_time, local_time):
        # Sometimes helpful diagnostic info
        self.info['reward_buffer.remote_time'] = remote_time
        self.info['reward_buffer.local_time'] = local_time

    def set_env_info(self, env_state):
        self._env_state = env_state

    def push_text(self, text):
        self.text.append(text)

    def push_done(self, done, info):
        # Consider yourself done whenever a reward crosses episode
        # boundaries.
        self.done = self.done or done
        if done:
            info['reward_buffer.done_time'] = time.time()
        self.push_info(info)

    def push_info(self, info):
        merge.merge_infos(self.info, info)

    def push(self, reward, done, info):
        # extra_logger.debug('[%s] RewardState: pushing reward %s to episode_id %s', self.label, reward, self._episode_id)
        self.count += 1
        self.reward += reward

        # Consider yourself done whenever a reward crosses episode
        # boundaries.
        self.push_done(done, info)

    def pop_text(self):
        text = self.text
        self.text = []
        return text

    def pop_info(self):
        info = self.info
        self.info = {}
        info['env.text'] = self.pop_text()
        if self._observation is not None:
            # Only used for the debugging gym-core envs with
            # rewarder_observation set.
            info['rewarder.observation'] = (self._observation, self._episode_id)
        info['env_status.episode_id'] = self._episode_id
        info['env_status.env_state'] = self._env_state
        return info

    def pop(self):
        info = self.pop_info()

        count = self.count
        reward = self.reward
        done = self.done

        self.count = 0
        self.reward = 0.
        self.done = False

        info['stats.reward.count'] = count
        extra_logger.debug('[%s] RewardState: popping reward %s from episode_id %s', self.label, reward, self._episode_id)
        return reward, done, info

    def set_observation(self, observation):
        self._observation = observation

# Buffers up incoming rewards
class RewardBuffer(object):
    def __init__(self, label):
        self.cv = threading.Condition()
        self.label = label

        self._current_episode_id = None
        self._reward_state = {}

        self._masked = True

        self._remote_env_state = None
        self._remote_env_id = None
        self._remote_episode_id = None
        self._remote_fps = None

    def reward_state(self, episode_id):
        try:
            return self._reward_state[episode_id]
        except:
            extra_logger.info('[%s] RewardBuffer: Creating new RewardState for episode_id=%s', self.label, episode_id)
            reward_state = self._reward_state[episode_id] = RewardState(self.label, episode_id)
            if self._current_episode_id is None and not self._masked:
                extra_logger.info('[%s] RewardBuffer advancing: No active episode, so activating episode_id=%s', self.label, episode_id)
                self._current_episode_id = episode_id
                self._drop_below(episode_id)
            if episode_id is not None and not self._masked:
                # If we're masked we'll be dropping everything below the reset ID anyway
                valid = self._valid_ids()
                for id in valid:
                    if id == episode_id:
                        continue
                    state = self._reward_state[id]
                    if state.done:
                        continue
                    extra_logger.info('[%s] RewardBuffer received message for episode_id=%s but no done=True message for %s. Artificially marking %s as done=True.', self.label, episode_id, id, id)
                    state.push_done(True, {'env_status.artificial.done': True})

            return reward_state

    def set_env_info(self, env_state, env_id, episode_id, fps):
        with self.cv:
            if self._remote_env_state is not None:
                extra_logger.info('[%s] RewardBuffer changing env_state: %s (env_id=%s) -> %s (env_id=%s) (episode_id: %s->%s, fps=%s, masked=%s, current_episode_id=%s)', self.label, self._remote_env_state, self._remote_env_id, env_state, env_id, self._remote_episode_id, episode_id, fps, self._masked, self._current_episode_id)
            else:
                extra_logger.info('[%s] RewardBuffer: Initial env_state: %s (env_id=%s) (episode_id: %s, fps=%s, masked=%s, current_episode_id=%s)', self.label, env_state, env_id, episode_id, fps, self._masked, self._current_episode_id)

            self._remote_env_state = env_state
            self._remote_env_id = env_id
            self._remote_episode_id = episode_id
            self._remote_fps = fps

            self.reward_state(episode_id).set_env_info(env_state)

    def set_observation(self, episode_id, observation):
        with self.cv:
            self.reward_state(episode_id).set_observation(observation)

    def push_time(self, episode_id, remote_time, local_time):
        with self.cv:
            self.reward_state(episode_id).push_time(remote_time, local_time)

    def push_text(self, episode_id, text):
        with self.cv:
            self.reward_state(episode_id).push_text(text)
            self.cv.notifyAll()

    def push_info(self, episode_id, info):
        # Just send some info
        with self.cv:
            self.reward_state(episode_id).push_info(info)

    def push(self, episode_id, reward, done, info):
        with self.cv:
            self.reward_state(episode_id).push(reward, done, info)
            self.cv.notifyAll()

    def pop(self, peek=False):
        with self.cv:
            self.cv.notifyAll()
            if peek:
                # This happens when a higher layer wants to poll for
                # new observations being ready, but doesn't want to
                # pop any rewards.
                max_id = self._max_id()
                reward_state = self.reward_state(self._current_episode_id)
                peek_state = self.reward_state(max_id)
                peek_id = peek_state._episode_id
                peek_state = peek_state._env_state
                if self._masked:
                    assert reward_state._episode_id is None
                    assert reward_state._env_state is None
                    peek_id = None
                    peek_state = None
                return 0, False, {
                    'peek': True,

                    'env_status.episode_id': reward_state._episode_id,
                    'env_status.env_state': reward_state._env_state,

                    'env_status.peek.episode_id': peek_id,
                    'env_status.peek.env_state': peek_state,
                }

            reward, done, info = self.reward_state(self._current_episode_id).pop()
            if done:
                # We return the *observation* from the new,
                # reward/done from the old, and a merged info with
                # keys from the new taking precedence.
                self._advance()

                new_state = self.reward_state(self._current_episode_id)
                try:
                    info['env_status.complete.episode_id'] = info['env_status.episode_id']
                except KeyError:
                    pass
                try:
                    info['env_status.complete.env_state'] = info['env_status.env_state']
                except KeyError:
                    pass
                info['env_status.episode_id'] = new_state._episode_id
                info['env_status.env_state'] = new_state._env_state
                new_text = self.reward_state(self._current_episode_id).pop_text()
                if len(info['env.text']) > 0:
                    extra_logger.info('[%s] RewardBuffer dropping env.text for completed episode %s: %s', self.label, info['env_status.episode_id'], info['env.text'])
                info['env.text'] = new_text
            return reward, done, info

    def mask(self):
        with self.cv:
            extra_logger.info('[%s] RewardBuffer advancing: masking until reset completes; setting current_episode_id=None', self.label)
            self._masked = True
            self._current_episode_id = None

    def reset(self, episode_id):
        with self.cv:
            extra_logger.info('[%s] RewardBuffer advancing: unmasking after explicit reset: episode_id=%s', self.label, episode_id)
            self._masked = False
            self._drop_below(episode_id, quiet=True)
            self._current_episode_id = episode_id
            self.push_info(episode_id, {'env_status.reset.episode_id': episode_id})

    def _max_id(self):
        valid_ids = self._valid_ids()
        if len(valid_ids) > 0:
            parsed = max(env_status.parse_episode_id(k) for k in valid_ids)
            return env_status.generate_episode_id(parsed)
        else:
            return None

    def _valid_ids(self):
        return [r for r in self._reward_state.keys() if r is not None]

    def _advance(self):
        completed_episode_id = self._current_episode_id
        del self._reward_state[completed_episode_id]

        if None in self._reward_state:
            extra_logger.warn('[%s] WARNING: RewardBuffer: while advancing from %s, None was in reward state: %s', self.label, completed_episode_id, self._reward_state)

        max_id = self._max_id()
        if max_id is not None:
            self._current_episode_id = max_id
            if env_status.compare_ids(completed_episode_id, self._current_episode_id) >= 0:
                extra_logger.info("[%s] RewardBuffer advancing: setting episode_id=None until new data received. Rare condition reached where message for old environment received after new one: completed_episode_id=%r self._current_episode_id=%r (%r). This is ok, but something we may want to fix in the future", self.label, completed_episode_id, self._current_episode_id, self._reward_state)
                self._current_episode_id = None
            else:
                extra_logger.info('[%s] RewardBuffer advancing: has data for next episode: %s->%s', self.label, completed_episode_id, self._current_episode_id)
                self._drop_below(self._current_episode_id)
        else:
            extra_logger.info('[%s] RewardBuffer advancing: setting episode_id=None until new data received (was episode_id=%s)', self.label, completed_episode_id)
            self._current_episode_id = None

    def _drop_below(self, episode_id, quiet=False):
        dropped = set()
        for stored_id in self._reward_state:
            if env_status.compare_ids(stored_id, episode_id) < 0:
                dropped.add(stored_id)

        if len(dropped) > 0:
            if quiet:
                log = extra_logger.debug
            else:
                log = extra_logger.info
            log('[%s] RewardBuffer: dropping stale episode data: dropped=%s episode_id=%s', self.label, dropped, episode_id)
        for stored_id in dropped:
            del self._reward_state[stored_id]

    def wait_for_step(self, error_buffer=None, timeout=None):
        # TODO: this might be cleaner using channels
        with self.cv:
            start = time.time()
            while True:
                if self.count != 0:
                    return
                elif timeout is not None and time.time() - start > timeout:
                    raise error.Error('No rewards received in {}s'.format(timeout))

                if error_buffer:
                    error_buffer.check()

                self.cv.wait(timeout=0.5)



from universe import spaces
from universe.envs.vnc_core_env import key
import logging

logger = logging.getLogger(__name__)


class AtariKeyState(object):
    """
    Converts from VNCEvents to an Atari-v0 action index

    Since spaces.KeyEvent only give you a diff of a keyboard, we need to persist the total state of the keyboard to
    convert from VNCEvents to an action index
    """
    def __init__(self, env):
        self._translator = AtariTranslator(env)
        self._down_keysyms = set()  # Assumes that your env starts with no keys pressed down

    def apply_vnc_actions(self, vnc_actions):
        """
        Play a list of vnc_actions forward over the current keysyms state

        NOTE: Since we are squashing a set of diffs into a single keyboard state, some information may be lost.
        For example if the Z key is down, then we receive [(Z-up), (Z-down)], the output will not reflect any change in Z
        You can make each frame shorter to offset this effect.
        """
        for event in vnc_actions:
            if isinstance(event, spaces.KeyEvent):
                if event.down:
                    self._down_keysyms.add(event.key)
                else:
                    self._down_keysyms.discard(event.key)

        logger.debug("AtariKeyState._down_keysyms: {}".format(self._down_keysyms))

    def to_keysyms(self):
        """Returns the current state as keysyms"""
        return list(self._down_keysyms)

    def to_index(self):
        """Returns the current state as an index"""
        return self._translator.keysyms_to_index(self.to_keysyms())


class AtariTranslator(object):
    """Translates Atari actions to and from various formats"""
    _all_keysyms = [key.UP, key.DOWN, key.LEFT, key.RIGHT, key.Z]

    def __init__(self, env):
        # e.g. {0: 'NOOP', 1: 'FIRE', 2: 'RIGHT', 3: 'LEFT', 4: 'RIGHTFIRE', 5: 'LEFTFIRE'}
        self._index_to_name_ = {}
        # e.g. {'RIGHT': 2, 'FIRE': 1, 'RIGHTFIRE': 4, 'LEFTFIRE': 5, 'NOOP': 0, 'LEFT': 3}
        self._name_to_index_ = {}

        for i, meaning in enumerate(env.unwrapped.get_action_meanings()):
            self._name_to_index_[meaning] = i
            self._index_to_name_[i] = meaning

    def keysyms_to_vnc_actions(self, keysyms):
        actions = []
        keysyms = set(keysyms)
        for keysym in self._all_keysyms:
            down = keysym in keysyms
            actions.append(spaces.KeyEvent(keysym, down=down))
        return actions

    def keysyms_to_index(self, keysyms):
        name = self._keysyms_to_name(keysyms)
        return self._name_to_index(name)

    def index_to_keysyms(self, i):
        name = self._index_to_name(i)
        keysyms = []
        if 'UP' in name:
            keysyms.append(key.UP)
        if 'DOWN' in name:
            keysyms.append(key.DOWN)
        if 'LEFT' in name:
            keysyms.append(key.LEFT)
        if 'RIGHT' in name:
            keysyms.append(key.RIGHT)
        if 'FIRE' in name:
            keysyms.append(key.Z)
        return keysyms

    def _name_to_index(self, name):
        return self._name_to_index_.get(name, 0)

    def _index_to_name(self, i):
        return self._index_to_name_[i]

    def _keysyms_to_name(self, keysyms):
        keys = ''
        if key.UP in keysyms:
            keys += 'UP'
        if key.DOWN in keysyms:
            keys += 'DOWN'
        if key.LEFT in keysyms:
            keys += 'LEFT'
        if key.RIGHT in keysyms:
            keys += 'RIGHT'
        if key.Z in keysyms:
            keys += 'FIRE'
        return keys

class CartPoleTranslator(object):
    def __init__(self, env):
        pass

    def keysyms_to_vnc_actions(self, keysyms):
        down = key.LEFT in keysyms
        return [spaces.KeyEvent(key.LEFT, down=down)]

    def keysyms_to_index(self, keys):
        if key.LEFT in keys:
            return 0
        else:
            return 1

    def index_to_keysyms(self, i):
        if i == 0:
            return [key.LEFT]
        else:
            return []



from __future__ import absolute_import
from __future__ import unicode_literals

import signal


class ShutdownException(Exception):
    pass


def shutdown(signal, frame):
    raise ShutdownException()


def set_signal_handler(handler):
    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, handler)


def set_signal_handler_to_shutdown():
    set_signal_handler(shutdown)



# -*- coding: utf-8 -*-
import logging

import six
import numpy as np

logger = logging.getLogger(__name__)

# We log these with logger, which in py2 chokes on unicode
def fmt_plusminus(mean, dev):
    if six.PY3:
        return mean + '±' + dev
    else:
        # Logging unicode in py2 is asking for trouble
        return mean + '+-' + dev

def compute_timestamps_pair_max(time_m_2, flat=True):
    if flat:
        # Ignore empty inputs, which happens when environments are resetting.
        time_m_2 = [[x for x in time_m_2 if len(x)]]

    if len(time_m_2) == 0:
        return None, None

    # We concatenate the (min, max) lags from a variety of runs. Those
    # runs may have different lengths.
    time_m_2 = [np.array(m) for m in time_m_2]

    timestamp_m = []
    data_m = []
    for m in time_m_2:
        if len(m) > 0:
            timestamp, data = compute_timestamps_sigma(m[:, 1])
            timestamp_m.append(timestamp)
            data_m.append(data)
        else:
            timestamp_m.append(None)
            data_m.append({})
    return timestamp_m, data_m

def display_timestamps_pair_compact(time_m_2):
    """Takes a list of the following form: [(a1, b1), (a2, b2), ...] and
    returns a string a_mean-b_mean, flooring out at 0.
    """
    if len(time_m_2) == 0:
        return '(empty)'

    time_m_2 = np.array(time_m_2)

    low = time_m_2[:, 0].mean()
    high = time_m_2[:, 1].mean()

    low = max(low, 0)

    # Not sure if this'll always be true, and not worth crashing over
    if high < 0:
        logger.warn('Harmless warning: upper-bound on clock skew is negative: (%s, %s). Please let Greg know about this.', low, high)

    return '{}-{}'.format(display_timestamp(low), display_timestamp(high))

def display_timestamps_pair(time_m_2):
    """Takes a list of the following form: [(a1, b1), (a2, b2), ...] and
    returns a string (a_mean+/-a_error, b_mean+/-b_error).
    """
    if len(time_m_2) == 0:
        return '(empty)'

    time_m_2 = np.array(time_m_2)
    return '({}, {})'.format(
        display_timestamps(time_m_2[:, 0]),
        display_timestamps(time_m_2[:, 1]),
    )

def compute_timestamps_sigma_n(time_m):
    timestamp_m = []
    data_m = []

    for t in time_m:
        timestamp, data = compute_timestamps(t)
        timestamp_m.append(timestamp)
        data_m.append(data)

    return timestamp_m, data_m

def compute_timestamps_sigma(time_m):
    if len(time_m) == 0:
        return None, {}

    mean = np.mean(time_m)
    std = standard_error(time_m)
    scale, units = pick_time_units(mean)
    return fmt_plusminus('{:.2f}{}'.format(mean * scale, units), '{:.2f}{}'.format(std * scale, units)), {'mean': mean}

def display_timestamps(time_m):
    res, _ = compute_timestamps(time_m)
    if res is None:
        return '(empty)'
    else:
        return res

def compute_timestamps(time_m):
    if len(time_m) == 0:
        return None, {}

    mean = np.mean(time_m)
    std = standard_error(time_m)
    return fmt_plusminus(display_timestamp(mean), display_timestamp(std)), {'mean': mean}

def display_timestamps_n(time_m):
    # concatenate all the n's timesteps together, then display_timestamps on it
    return display_timestamps(np.concatenate(time_m))

def standard_error(ary, axis=0):
    if len(ary) > 1:
        return np.std(ary, axis=axis) / np.sqrt(len(ary) - 1)
    else:
        return np.std(ary, axis=axis)

def display_timestamp(time):
    assert not isinstance(time, np.ndarray), 'Invalid scalar: {}'.format(time)
    scale, units = pick_time_units(time)
    return '{:.2f}{}'.format(time * scale, units)

def pick_time_units(time):
    assert not isinstance(time, np.ndarray), 'Invalid scalar: {}'.format(time)
    if abs(time) < 1:
        return 1000, 'ms'
    else:
        return 1, 's'
